// -------------------------------------------------------------------------------- 
//!  @file  pce.c                                                                   
//!  PC エンジンパッドソースファイル                                                
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "config.h"
#include "pad.h"
#include "pce.h"

// PORT RESET : SEL = 1 && CLR = 0 -> 1
// NEXT PORT  : CLR = 0 && SEL = 0 -> 1
// PAD6????   : SEL = 0 && CLR = 1 -> 0


#ifdef  PAD_PCE_CFG_PIN_SEL

// -------------------------------------------------------------------------------- 
//!  1usec 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC wait1us( void )
{
	volatile int	index ;


	for(index=0; index<11; index++) {
		asm volatile("nop") ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ポート設定                                                                     
//!                                                                                 
//!  @param [in] gpio   GPIO 番号                                                   
//!  @param [in] value  値                                                          
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static inline void setPort(uint gpio, bool value)
{
	gpio_put(gpio, value) ;

	wait1us() ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void pad_pce_init( void )
{
	gpio_init( PAD_PCE_CFG_PIN_SEL ) ;
	gpio_init( PAD_PCE_CFG_PIN_CLR ) ;
	gpio_init( PAD_PCE_CFG_PIN_D0 ) ;
	gpio_init( PAD_PCE_CFG_PIN_D1 ) ;
	gpio_init( PAD_PCE_CFG_PIN_D2 ) ;
	gpio_init( PAD_PCE_CFG_PIN_D3 ) ;

	gpio_disable_pulls( PAD_PCE_CFG_PIN_SEL ) ;
	gpio_disable_pulls( PAD_PCE_CFG_PIN_CLR ) ;
	gpio_pull_down( PAD_PCE_CFG_PIN_D0 ) ;
	gpio_pull_down( PAD_PCE_CFG_PIN_D1 ) ;
	gpio_pull_down( PAD_PCE_CFG_PIN_D2 ) ;
	gpio_pull_down( PAD_PCE_CFG_PIN_D3 ) ;

	gpio_put(PAD_PCE_CFG_PIN_SEL,     1) ;
	gpio_put(PAD_PCE_CFG_PIN_CLR,     0) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_SEL, GPIO_OUT) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_CLR, GPIO_OUT) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_D0,  GPIO_IN) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_D1,  GPIO_IN) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_D2,  GPIO_IN) ;
	gpio_set_dir(PAD_PCE_CFG_PIN_D3,  GPIO_IN) ;
}


// -------------------------------------------------------------------------------- 
//!  パッド状態を取得                                                               
//!                                                                                 
//!  @param [in,out] p_padInfo  パッド情報                                          
//!  @return                    なし                                                
// -------------------------------------------------------------------------------- 
void RAMFUNC pad_pce_get( PADINFO *p_padInfo )
{
	uint8_t		status[4] ;

	// リセット
	gpio_put(PAD_PCE_CFG_PIN_SEL, 1) ;
	setPort(PAD_PCE_CFG_PIN_CLR,  0) ;
	setPort(PAD_PCE_CFG_PIN_CLR,  1) ;

	// ｜左｜下｜右｜上｜ or ｜０｜０｜０｜０｜
	setPort(PAD_PCE_CFG_PIN_CLR, 0) ;
    status[0] = (~gpio_get_all() >> PAD_PCE_CFG_PIN_D0) & 15 ;

	// ｜Ｒ｜Ｓ｜Ⅱ｜Ⅰ| or ｜Ⅵ｜Ⅴ｜Ⅳ｜Ⅲ｜
	setPort(PAD_PCE_CFG_PIN_SEL, 0) ;
	status[1] = (~gpio_get_all() >> PAD_PCE_CFG_PIN_D0) & 15 ;

	if( status[0] != 15 ) {			// ２ボタン

		p_padInfo->type = PAD_TYPE_PCE ;

		// Ⅱ -> Ⅲ
		status[2] = (status[1] >> 1) & 1 ;

		// Ⅰ -> Ⅱ
		status[1] = (status[1] & 0x0c) | ((status[1] & 1) << 1) ;

	} else {

		if( status[1] == 15 ) {		// 未接続

			p_padInfo->type = PAD_TYPE_UNKNOWN ;
			status[0]       = 0 ;
			status[1]       = 0 ;
			status[2]       = 0 ;

		} else {					// ６ボタン

			p_padInfo->type = PAD_TYPE_PCE_6B ;

			// ｜Ⅵ｜Ⅴ｜Ⅳ｜Ⅲ｜
			status[2] = status[1] ;

			// 切り替え
			setPort(PAD_PCE_CFG_PIN_CLR, 1) ;
			setPort(PAD_PCE_CFG_PIN_SEL, 1) ;
			setPort(PAD_PCE_CFG_PIN_CLR, 0) ;

			// ｜Ｒ｜Ｓ｜Ⅱ｜Ⅰ|
			status[0] = (~gpio_get_all() >> PAD_PCE_CFG_PIN_D0) & 15 ;

			// ｜左｜下｜右｜上｜
			setPort(PAD_PCE_CFG_PIN_SEL, 0) ;
		    status[1] = (~gpio_get_all() >> PAD_PCE_CFG_PIN_D0) & 15 ;

			// 連射速度調整
			p_padInfo->reserved ^= 1 ;
			if( p_padInfo->reserved & 1 ) {
				setPort(PAD_PCE_CFG_PIN_CLR, 1) ;
				setPort(PAD_PCE_CFG_PIN_CLR, 0) ;
				setPort(PAD_PCE_CFG_PIN_CLR, 1) ;
				setPort(PAD_PCE_CFG_PIN_CLR, 0) ;
			}

		}

	}

	// 次パッド
	setPort(PAD_PCE_CFG_PIN_SEL, 0) ;
	setPort(PAD_PCE_CFG_PIN_SEL, 1) ;

	p_padInfo->status = ((uint16_t )((status[0] >> 3) & 1) << PAD_LEFT_)   |
	                    ((uint16_t )((status[0] >> 1) & 1) << PAD_RIGHT_)  |
	                    ((uint16_t )((status[0] >> 0) & 1) << PAD_UP_)     |
	                    ((uint16_t )((status[0] >> 2) & 1) << PAD_DOWN_)   |
	                    ((uint16_t )((status[1] >> 2) & 1) << PAD_SELECT_) |
	                    ((uint16_t )((status[1] >> 3) & 1) << PAD_START_)  |
	                    ((uint16_t )((status[2] >> 0) & 1) << (PAD_BTN_+0)) |
	                    ((uint16_t )((status[1] >> 1) & 1) << (PAD_BTN_+1)) |
	                    ((uint16_t )((status[1] >> 0) & 1) << (PAD_BTN_+2)) |
	                    ((uint16_t )((status[2] >> 1) & 1) << (PAD_BTN_+3)) |
	                    ((uint16_t )((status[2] >> 2) & 1) << (PAD_BTN_+4)) |
	                    ((uint16_t )((status[2] >> 3) & 1) << (PAD_BTN_+5)) ;
}

#endif//PAD_PCE_CFG_PIN_SEL
