// -------------------------------------------------------------------------------- 
//!  @file  namcot163.h                                                             
//!  NAMCOT-163 音源 PCM データジェネレータヘッダファイル                           
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#ifndef __PCM__NAMCOT163_H__INCLUDED__
#define __PCM__NAMCOT163_H__INCLUDED__

// レジスタ
//   [0] |----VVVV|------FF|FFFFFFFF|FFFFFFFF|
//   [1] |--------|--------|--LLLLLL|AAAAAAAA|
//   [2] |--------|PPPPPPPP|PPPPPPPP|PPPPPPPP|

	enum {
		PCMNAMCOT163_WAVE_SIZE		=	128,						//!< 波形サイズ
		PCMNAMCOT163_CHANNEL_COUNT	=	8							//!< チャンネル数
	} ;

	//! チャンネル情報
	typedef union {
		uint32_t					reg[3] ;						//!< レジスタ
		uint8_t						regb[12] ;						//!< レジスタ（バイトアクセス）
	} PCMNAMCOT163CHANNELINFO ;

	//! NAMCOT-163 情報
	typedef struct {
		uint8_t						wave[PCMNAMCOT163_WAVE_SIZE] ;	//!< 波形 RAM
		PCMNAMCOT163CHANNELINFO		channelInfo[PCMNAMCOT163_CHANNEL_COUNT] ;
																	//!< チャンネル情報
		uint8_t						channelIndex ;					//!< チャンネルインデックス
		uint8_t						channelMax ;					//!< チャンネル最大数
		uint8_t						enable ;						//!< チャンネル有効フラグ
	} PCMNAMCOT163INFO ;


extern void            pcm_namcot163_reset( PCMNAMCOT163INFO *p ) ;
extern int16_t RAMFUNC pcm_namcot163_update( PCMNAMCOT163INFO *p ) ;


// -------------------------------------------------------------------------------- 
//!  有効・無効                                                                     
//!                                                                                 
//!  @param [out] p       NAMCOT-163 情報                                           
//!  @param [in]  enable  チャンネル有効ビット                                      
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_enable(PCMNAMCOT163INFO *p, bool enable)
{
	p->enable = enable ;
}


// -------------------------------------------------------------------------------- 
//!  チャンネル数設定                                                               
//!                                                                                 
//!  @param [out] p      NAMCOT-163 情報                                            
//!  @param [in]  count  チャンネル数                                               
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_setChannelCount(PCMNAMCOT163INFO *p, int count)
{
	if( (uint )(count-1) >= 8  ) {
		return ;
	}

	p->channelMax = count ;
}


// -------------------------------------------------------------------------------- 
//!  波形データを波形 RAM へコピー                                                  
//!                                                                                 
//!  @param [out] p       NAMCOT-163 情報                                           
//!  @param [in]  offset  開始オフセット                                            
//!  @param [in]  p_wave  波形データアドレス                                        
//!  @param [in]  size    波形データサイズ                                          
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_copyWaveData(PCMNAMCOT163INFO *p, uint8_t offset, uint8_t const *p_wave, uint8_t size)
{
	memcpy(&p->wave[offset], &p_wave[0], size) ;
}


// -------------------------------------------------------------------------------- 
//!  レジスタ設定                                                                   
//!                                                                                 
//!  @param [out] p       NAMCOT-163 情報                                           
//!  @param [in]  channel チャンネル番号                                            
//!  @param [in]  value   レジスタ値                                                
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_setReg(PCMNAMCOT163INFO *p, uint8_t channel, uint8_t index, uint32_t value)
{
	p->channelInfo[channel].reg[index] = value ;
}


// -------------------------------------------------------------------------------- 
//!  周波数と音量設定                                                               
//!                                                                                 
//!  @param [out] p       NAMCOT-163 情報                                           
//!  @param [in]  channel チャンネル番号                                            
//!  @param [in]  freq    周波数                                                    
//!  @param [in]  volume  音量                                                      
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_setFreqVolume(PCMNAMCOT163INFO *p, uint8_t channel, uint32_t freq, uint8_t volume)
{
	p->channelInfo[channel].reg[0] = ((uint32_t )volume << 24) | freq ;
}


// -------------------------------------------------------------------------------- 
//!  アドレス設定                                                                   
//!                                                                                 
//!  @param [out] p        NAMCOT-163 情報                                          
//!  @param [in]  channel  チャンネル番号                                           
//!  @param [in]  address  アドレス                                                 
//!  @param [in]  length   長さ                                                     
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_setAddress(PCMNAMCOT163INFO *p, uint8_t channel, uint8_t address, uint8_t length)
{
	p->channelInfo[channel].reg[1] = ((uint32_t )length << 8) | address ;
}


// -------------------------------------------------------------------------------- 
//!  位相設定                                                                       
//!                                                                                 
//!  @param [out] p        NAMCOT-163 情報                                          
//!  @param [in]  channel  チャンネル番号                                           
//!  @param [in]  phase    位相                                                     
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void pcm_namcot163_setPhase(PCMNAMCOT163INFO *p, uint8_t channel, uint32_t phase)
{
	p->channelInfo[channel].reg[2] = phase ;
}


#endif//__PCM__NAMCOT163_H__INCLUDED__
