// -------------------------------------------------------------------------------- 
//!  @file  sn76489.c                                                               
//!  SN76489 音源 PWM 出力ソースファイル                                            
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "hardware/pwm.h"
#include "config.h"
#include "sn76489.h"


#ifdef  PWMSN76489_CFG_BASECLK

	enum {
		CLK							=	125000000,					//!< クロック
		CLKDIV						=	125,						//!< クロック分周
		NOISE_TIMER_US				=	36*12/2,					//!< ノイズタイマーマイクロ秒
		DUTY_ONE					=	64							//!< デューティ比固定小数１
	} ;

	//! トーン
	typedef struct {
		uint8_t						slice ;							//!< PWM スライス
		uint8_t						channel ;						//!< PWM チャンネル
		uint16_t					value ;							//!< 設定値
	} TONE ;

	//! サウンド情報
	typedef struct {
		repeating_timer_t			timer ;							//!< 周期タイマー
		TONE						tone[3] ;						//!< トーン情報

		//! ノイズ
		struct {
			uint16_t				value ;							//!< 設定値
			uint16_t				output ;						//!< 出力ビット
		} noise ;
	} SOUNDINFO ;


static SOUNDINFO					m_soundInfo ;					//!< サウンド情報

//<! 音量変換
static uint8_t						m_volume[]={
										56,		//  0
										55,		//  1
										54,		//  2
										53,		//  3
										51,		//  4
										48,		//  5
										45,		//  6
										41,		//  7
										37,		//  8
										32,		//  9
										28,		// 10
										22,		// 11
										17,		// 12
										11,		// 13
										 5,		// 14
										 0,		// 15
									} ;


// -------------------------------------------------------------------------------- 
//!  周波数カウントから周波数を取得                                                 
//!                                                                                 
//!  @param [in] count  周波数カウント                                              
//!  @return            周波数                                                      
// -------------------------------------------------------------------------------- 
static inline uint32_t get_freq( uint count )
{
	return  ((CLK / CLKDIV) * (count + 1)) / PWMSN76489_CFG_BASECLK ;
}


// -------------------------------------------------------------------------------- 
//!  タイマー割り込みサービスルーチン                                               
//!                                                                                 
//!  @retval true   タイマー継続                                                    
//!  @retval false  タイマー停止                                                    
// -------------------------------------------------------------------------------- 
static bool isr_timer( struct repeating_timer *p_timer )
{
	if( m_soundInfo.noise.value & (1<<10) ) {	// ホワイトノイズ
		m_soundInfo.noise.output = (m_soundInfo.noise.output >> 1) | (((m_soundInfo.noise.output << 12) ^ (m_soundInfo.noise.output << 15)) & 0x8000) ;
	} else {									// 同期ノイズ
		m_soundInfo.noise.output = (m_soundInfo.noise.output >> 1) |                                      (m_soundInfo.noise.output << 15) ;
	}

	if( (m_soundInfo.noise.value>>12) == 15 ) {		// 消音
		gpio_put(PWMSN76489_CFG_PIN_NOISE, 0) ;
	} else {
		gpio_put(PWMSN76489_CFG_PIN_NOISE, m_soundInfo.noise.output&1) ;
	}

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  動作開始                                                                       
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
void pwm_sn76489_start( void )
{
	TONE	*p_tone ;
	int		index ;

	// 初期化
	m_soundInfo.tone[0].slice   = PWMSN76489_CFG_TONE0_SLICE ;
	m_soundInfo.tone[0].channel = PWMSN76489_CFG_TONE0_CHANNEL ;
	m_soundInfo.tone[1].slice   = PWMSN76489_CFG_TONE1_SLICE ;
	m_soundInfo.tone[1].channel = PWMSN76489_CFG_TONE1_CHANNEL ;
	m_soundInfo.tone[2].slice   = PWMSN76489_CFG_TONE2_SLICE ;
	m_soundInfo.tone[2].channel = PWMSN76489_CFG_TONE2_CHANNEL ;
	m_soundInfo.noise.value     = 15 << 12 ;
	m_soundInfo.noise.output    = 0x8000 ;

	gpio_set_function(PWMSN76489_CFG_PIN_TONE0, GPIO_FUNC_PWM) ;
	gpio_set_function(PWMSN76489_CFG_PIN_TONE1, GPIO_FUNC_PWM) ;
	gpio_set_function(PWMSN76489_CFG_PIN_TONE2, GPIO_FUNC_PWM) ;

	// トーン
	p_tone = &m_soundInfo.tone[0] ;
	for(index=0; index<3; index++,p_tone++) {
		p_tone->value = 0xffff ;
		pwm_set_clkdiv_int_frac(p_tone->slice, CLKDIV, 0) ;
		pwm_set_wrap(p_tone->slice, get_freq( 4095 )) ;
		pwm_set_chan_level(p_tone->slice, p_tone->channel, 0) ;
	}

	// ノイズ
	gpio_init( PWMSN76489_CFG_PIN_NOISE ) ;
	gpio_set_dir(PWMSN76489_CFG_PIN_NOISE, GPIO_OUT) ;
	gpio_put(PWMSN76489_CFG_PIN_NOISE, 0) ;

	add_repeating_timer_us(-NOISE_TIMER_US, isr_timer, NULL, &m_soundInfo.timer) ;

	pwm_set_mask_enabled( pwm_hw->en | ((1<<PWMSN76489_CFG_TONE2_SLICE)|(1<<PWMSN76489_CFG_TONE1_SLICE)|(1<<PWMSN76489_CFG_TONE0_SLICE)) ) ;
}


// -------------------------------------------------------------------------------- 
//!  動作停止                                                                       
//!                                                                                 
//!  @return  0                                                                     
// -------------------------------------------------------------------------------- 
int sound_stop( void )
{
	pwm_set_mask_enabled( pwm_hw->en & ~((1<<PWMSN76489_CFG_TONE2_SLICE) | (1<<PWMSN76489_CFG_TONE1_SLICE) | (1<<PWMSN76489_CFG_TONE0_SLICE)) ) ;

	cancel_repeating_timer( &m_soundInfo.timer ) ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  トーン設定                                                                     
//!                                                                                 
//!  @param [in] tone    トーン番号                                                 
//!  @param [in] value   周波数カウント＋音量                                       
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
void sound_setTone(int tone, uint16_t value)
{
	TONE		*p_tone ;
	uint32_t	freq ;


	p_tone = &m_soundInfo.tone[tone] ;
	if( p_tone->value == value ) {
		return ;
	}
	p_tone->value = value ;

	freq = get_freq( value & 4095 ) ;

	pwm_set_wrap(p_tone->slice, freq) ;
	pwm_set_chan_level(p_tone->slice, p_tone->channel, (freq*m_volume[value>>12])/DUTY_ONE) ;
}


// -------------------------------------------------------------------------------- 
//!  ノイズ設定                                                                     
//!                                                                                 
//!  @param [in] value  ノイズ種類＋音量                                            
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
void sound_setNoise( uint16_t value )
{
	if( m_soundInfo.noise.value == value ) {
		return ;
	}

	if( (m_soundInfo.noise.value^value) & (1<<10) ) {
		m_soundInfo.noise.output = 0x8000 ;
	}

	m_soundInfo.noise.value = value ;
}


// -------------------------------------------------------------------------------- 
//!  レジスタ設定                                                                   
//!                                                                                 
//!  @param [in] channel  チャンネル番号                                            
//!  @param [in] value    レジスタ値                                                
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
void RAMFUNC pwm_sn76489_setReg(uint8_t channel, uint16_t value)
{
	if( channel < 3 ) {

		if( !(value&(1<<11)) ) {
			sound_setTone(channel, value) ;
		}

	} else {

		if( value & (1<<11) ) {
			sound_setNoise( value ) ;
		}

	}
}

#endif//PWMSN76489_CFG_BASECLK
