// -------------------------------------------------------------------------------- 
//!  @file  vga.c                                                                   
//!  VGA ソースファイル                                                             
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include <pico/stdlib.h>
#include <pico/multicore.h>
#include <hardware/irq.h>
#include <hardware/interp.h>
#include "config.h"
#include "pad/pad.h"
#include "vga.h"
#include "vga.pio.h"

//! パフォーマンス計測ピン番号（デバッグ用）
//#define DEBUG_PERFORMANCE			16

// アセンブラコード使用
#define USEASM

#if     (VGA_CFG_PIO == 0)
#define PIO							pio0							//!< PIO オブジェクト
#define PIO_IRQ						PIO0_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO0_TX0					//!< PIO データリクエスト
#else
#define PIO							pio1							//!< PIO オブジェクト
#define PIO_IRQ						PIO1_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO1_TX0					//!< PIO データリクエスト
#endif/*(VGA_CFG_PIO == 0)*/

#if     (VGA_CFG_DMA == 0)
#define DMA_IRQ						DMA_IRQ_0						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints0							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq0_enabled	//!< DMA 有効処理
#else
#define DMA_IRQ						DMA_IRQ_1						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints1							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq1_enabled	//!< DMA 有効処理
#endif//(VGA_CFG_DMA == 0)

#define BIT_TEST(n, b)				(((uint32_t )(n) << (31-(b))) >> 31)


#ifdef  USEASM
#define DRAW_PATTERN(i, s, n)									\
				#i "	r7, %[pattern], #" #s "			\n"		\
			"	and		r7, %[pmask]					\n"		\
			"	beq		0f								\n"		\
			"	ldrh	r7, [%[p_palette], r7]			\n"		\
			"	strh	r7, [%[p_buff], #" #n "]		\n"		\
			"0:											\n"

#define DRAW_TXT_SH_OP(i, s, n)									\
				#i "	r7, %[pattern], #" #s "			\n"		\
			"	and		r7, %[pmask]					\n"		\
			"	beq		0f								\n"		\
			"	ldrh	r7, [%[p_palette], r7]			\n"		\
			"	b		1f								\n"		\
			"0:	ldrh	r7, [%[p_buff], #" #n "]		\n"		\
			"	and		r7, %[smask]					\n"		\
			"	lsr		r7, #1							\n"		\
			"1:	strh	r7, [%[p_buff], #" #n "]		\n"

#define DRAW_TXT_SH(i, s, n)									\
				#i "	r7, %[pattern], #" #s "			\n"		\
			"	and		r7, %[pmask]					\n"		\
			"	beq		0f								\n"		\
			"	ldrh	r7, [%[p_buff], #" #n "]		\n"		\
			"	and		r7, %[smask]					\n"		\
			"	lsr		r7, #1							\n"		\
			"	strh	r7, [%[p_buff], #" #n "]		\n"		\
			"0:											\n"

#define DRAW_TXT_OP(i, s, n)									\
				#i "	r7, %[pattern], #" #s "			\n"		\
			"	and		r7, %[pmask]					\n"		\
			"	ldrh	r7, [%[p_palette], r7]			\n"		\
			"	strh	r7, [%[p_buff], #" #n "]		\n"
#endif//USEASM

	//! PIO ステートマシン番号
	enum {
		SM_CLKGEN_HORZ,												//!< VGA 水平クロックジェネレータ
		SM_CLKGEN_VERT,												//!< VGA 垂直クロックジェネレータ
		SM_TRANS_PIXEL												//!< VGA 有効ピクセル転送
	} ;

	//! ステートマシン情報
	typedef struct {
		uint8_t						pin ;							//!< ピン開始番号
		uint8_t						count ;							//!< ピン数
		uint8_t						sideset ;						//!< サイドセットフラグ
		uint32_t					out ;							//!< ピン出力値
		struct pio_program			const *p_program ;				//!< プログラム
		pio_sm_config (*getConfig)( uint offset ) ;					//!< コンフィグ取得処理
		uint32_t					scratchY ;						//!< スクラッチレジスター Y 初期値
	} PIOSMINFO ;

	//! スプライト情報
	typedef struct sprinfo {
		struct sprinfo				*p_next ;						//!< 次のスプライト情報
		uint16_t					x ;								//!< ｘ座標
		uint16_t					y ;								//!< ｙ座標
		uint32_t					attr ;							//!< 属性
	} SPRINFO ;


VGAINFO								g_vgaInfo ;						//!< VGA 情報
static interp_config				m_config8x8 ;					//!<   8x8 画像用インターポーレーター設定
static interp_config				m_config16x16 ;					//!< 16x16 画像用インターポーレーター設定

//! PIO ステートマシン情報
static const PIOSMINFO				m_pioSmInfo[]={
										{ VGA_CFG_PIN_HSYNC, 1,                      1, 1, &vga_horz_program,  &vga_horz_program_get_default_config,  (((48+640+16)*VGA_CFG_DOTCLK-1-1)<<16) | ( 96    *VGA_CFG_DOTCLK-3-1) },
										{ VGA_CFG_PIN_VSYNC, 1,                      1, 1, &vga_vert_program,  &vga_vert_program_get_default_config,  ((480                         -1)<<16) | ( 33-2                   -1) },
										{ VGA_CFG_PIN_PIXEL, VGA_CFG_PIN_PIXELCOUNT, 0, 0, &vga_pixel_program, &vga_pixel_program_get_default_config, ((320/2                       -1)<<16) | ((96+48)*VGA_CFG_DOTCLK-4-1) }
									} ;


// -------------------------------------------------------------------------------- 
//!  PIO 割り込みサービスルーチン                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC isr_pio( void )
{
	dma_channel_abort( VGA_CFG_DMACH_PIXEL ) ;

	PIO->irq           = 1 << 1 ;
	g_vgaInfo.dispLine = 0 ;
	g_vgaInfo.drawLine = 1 ;
}


// -------------------------------------------------------------------------------- 
//!  DMA 割り込みサービスルーチン                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC isr_dma( void )
{
	if( dma_hw->DMA_INTS & (1<<VGA_CFG_DMACH_PIXEL) ) {
		dma_hw->DMA_INTS = (1<<VGA_CFG_DMACH_PIXEL) ;

		if( ++g_vgaInfo.dispLine < VGA_HEIGHT ) {

			dma_channel_configure(VGA_CFG_DMACH_PIXEL, &g_vgaInfo.dmaConfigPixel, &PIO->txf[SM_TRANS_PIXEL], &g_vgaInfo.lineBuff.page[(g_vgaInfo.dispLine/2)&1].data[0], (VGA_IMAGE_WIDTH*2)/4, true) ;

			if( !(g_vgaInfo.dispLine&1) ) {
				g_vgaInfo.drawLine = g_vgaInfo.dispLine / 2 + 1 ;
			}

		} else {

#ifdef  VGA_CFG_PAD
			pad_update() ;
#endif//VGA_CFG_PAD

		}
	}
}


// -------------------------------------------------------------------------------- 
//!  PIO ステートマシンを初期化                                                     
//!                                                                                 
//!  @param [in] sm           ステートマシンインデックス                            
//!  @param [in] p_pioSmInfo  ステートマシン情報アドレス                            
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
static void initPioSm(uint sm, const PIOSMINFO *p_pioSmInfo)
{
	pio_sm_config	config ;
	int				index ;
	uint			offset ;
	uint32_t		out ;

	// GPIO 初期化
	out = p_pioSmInfo->out ;
	for(index=0; index<p_pioSmInfo->count; index++,out>>=1) {
		pio_gpio_init(PIO, p_pioSmInfo->pin+index) ;
		gpio_disable_pulls( p_pioSmInfo->pin+index ) ;
		gpio_put(p_pioSmInfo->pin+index, out&1) ;
	}

	pio_sm_set_consecutive_pindirs(PIO, sm, p_pioSmInfo->pin, p_pioSmInfo->count, true) ;

	offset = pio_add_program(PIO, p_pioSmInfo->p_program) ;
	config = p_pioSmInfo->getConfig( offset ) ;

	sm_config_set_set_pins(&config, p_pioSmInfo->pin, p_pioSmInfo->count) ;
	sm_config_set_out_pins(&config, p_pioSmInfo->pin, p_pioSmInfo->count) ;

	if( p_pioSmInfo->sideset ) {
		sm_config_set_sideset_pins(&config, p_pioSmInfo->pin) ;
	}

	pio_sm_init(PIO, sm, offset, &config) ;

	// スクラッチレジスター Y 初期化
	pio_sm_put_blocking(PIO, sm, p_pioSmInfo->scratchY) ;
	pio_sm_exec_wait_blocking(PIO, sm, pio_encode_pull(false, false)) ;
	pio_sm_exec_wait_blocking(PIO, sm, pio_encode_mov(pio_y, pio_osr)) ;
}


// -------------------------------------------------------------------------------- 
//!  内部情報を初期化                                                               
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void init( void )
{
	interp_config	config ;

	// タイルデータ
	config = interp_default_config() ;
	interp_config_set_mask(&config, 1, 6) ;
	interp_set_config(interp0, 0, &config) ;
	interp0->base[0] = 2 ;

	// バッファ
	config = interp_default_config() ;
	interp_config_set_shift(&config, 31) ;
	interp_config_set_mask(&config, 31, 31) ;
	interp_config_set_add_raw(&config, true) ;
	interp_set_config(interp0, 1, &config) ;
	interp0->base[1] = 16 ;

	// パレット
	config = interp_default_config() ;
	interp_config_set_shift(&config, 23) ;
	interp_config_set_mask(&config, 5, 8) ;
	interp_set_config(interp1, 0, &config) ;

	// 8x8 パターン
	m_config8x8 = interp_default_config() ;
	interp_config_set_shift(&m_config8x8, 11) ;
	interp_config_set_mask(&m_config8x8, 5, 14) ;
	interp_config_set_cross_input(&m_config8x8, true) ;

	// 16x16 パターン
	m_config16x16 = interp_default_config() ;
	interp_config_set_shift(&m_config16x16, 9) ;
	interp_config_set_mask(&m_config16x16, 7, 16) ;
	interp_config_set_cross_input(&m_config16x16, true) ;

	// パッド開始
#ifdef  VGA_CFG_PAD
	pad_start() ;
#endif//VGA_CFG_PAD

	// ステートマシン
	initPioSm(SM_CLKGEN_HORZ, &m_pioSmInfo[SM_CLKGEN_HORZ]) ;
	initPioSm(SM_CLKGEN_VERT, &m_pioSmInfo[SM_CLKGEN_VERT]) ;
	initPioSm(SM_TRANS_PIXEL, &m_pioSmInfo[SM_TRANS_PIXEL]) ;

	// 一部モニターで同期がふらつくためドライブストレングスを調整（現状が仕様を満たしていないと思われるので要調査）
	hw_write_masked(&padsbank0_hw->io[VGA_CFG_PIN_VSYNC], PADS_BANK0_GPIO0_DRIVE_VALUE_8MA<<PADS_BANK0_GPIO0_DRIVE_LSB, PADS_BANK0_GPIO0_DRIVE_BITS) ;
	hw_write_masked(&padsbank0_hw->io[VGA_CFG_PIN_HSYNC], PADS_BANK0_GPIO0_DRIVE_VALUE_8MA<<PADS_BANK0_GPIO0_DRIVE_LSB, PADS_BANK0_GPIO0_DRIVE_BITS) ;

	// ピクセルデータ転送開始割り込み設定
	hw_set_bits(&PIO->inte0, 1<<(8+1)) ;				// +1 == SM1
	PIO->irq = 1 << 1 ;

	irq_set_exclusive_handler(PIO_IRQ, isr_pio) ;
	irq_clear( PIO_IRQ ) ;
	irq_set_enabled(PIO_IRQ, true) ;

	// ピクセル転送用 DMA
	dma_channel_claim( VGA_CFG_DMACH_PIXEL ) ;
	g_vgaInfo.dmaConfigPixel = dma_channel_get_default_config( VGA_CFG_DMACH_PIXEL ) ;
	channel_config_set_dreq(&g_vgaInfo.dmaConfigPixel, PIO_DREQ+SM_TRANS_PIXEL) ;
	dma_enable(VGA_CFG_DMACH_PIXEL, true) ;

	irq_set_exclusive_handler(DMA_IRQ, isr_dma) ;
	irq_clear( DMA_IRQ ) ;
	irq_set_enabled(DMA_IRQ, true) ;

	// データ転送用 DMA
	dma_channel_claim( VGA_CFG_DMACH_DATA ) ;
	g_vgaInfo.dmaConfigData = dma_channel_get_default_config( VGA_CFG_DMACH_DATA ) ;
	channel_config_set_read_increment(&g_vgaInfo.dmaConfigData, false) ;
	channel_config_set_write_increment(&g_vgaInfo.dmaConfigData, true) ;

	// PIO 動作開始
	pio_set_sm_mask_enabled(PIO, (1<<SM_TRANS_PIXEL)|(1<<SM_CLKGEN_VERT)|(1<<SM_CLKGEN_HORZ), true) ;
}


// -------------------------------------------------------------------------------- 
//!  パターンデータを描画バッファへ描画                                             
//!                                                                                 
//!  @param [out] p_buff     描画バッファアドレス                                   
//!  @param [in]  p_palette  パレットデータアドレス                                 
//!  @param [in]  pattern    パターンデータ                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawPattern(uint16_t *p_buff, const uint8_t *p_palette, uint32_t pattern)
{
#ifdef  USEASM
	asm volatile (
		DRAW_PATTERN(lsl,  1,  0)
		DRAW_PATTERN(lsr,  3,  2)
		DRAW_PATTERN(lsr,  7,  4)
		DRAW_PATTERN(lsr, 11,  6)
		DRAW_PATTERN(lsr, 15,  8)
		DRAW_PATTERN(lsr, 19, 10)
		DRAW_PATTERN(lsr, 23, 12)
		DRAW_PATTERN(lsr, 27, 14)
		:
		: [p_buff] "l" (p_buff), [p_palette] "l" (p_palette), [pattern] "l" (pattern), [pmask] "l" (0x1e)
		: "r7"
	) ;
#else
	uint32_t	value ;


	value = (pattern <<  1) & 0x1e ;  if( value ) { p_buff[0] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >>  3) & 0x1e ;  if( value ) { p_buff[1] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >>  7) & 0x1e ;  if( value ) { p_buff[2] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 11) & 0x1e ;  if( value ) { p_buff[3] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 15) & 0x1e ;  if( value ) { p_buff[4] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 19) & 0x1e ;  if( value ) { p_buff[5] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 23) & 0x1e ;  if( value ) { p_buff[6] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 27) & 0x1e ;  if( value ) { p_buff[7] = *(uint16_t *)&p_palette[value] ; }
#endif//USEASM
}


// -------------------------------------------------------------------------------- 
//!  左右反転したパターンデータを描画バッファへ描画                                 
//!                                                                                 
//!  @param [out] p_buff     描画バッファアドレス                                   
//!  @param [in]  p_palette  パレットデータアドレス                                 
//!  @param [in]  pattern    パターンデータ                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawPatternFlip(uint16_t *p_buff, const uint8_t *p_palette, uint32_t pattern)
{
#ifdef  USEASM
	asm volatile (
		DRAW_PATTERN(lsl,  1, 14)
		DRAW_PATTERN(lsr,  3, 12)
		DRAW_PATTERN(lsr,  7, 10)
		DRAW_PATTERN(lsr, 11,  8)
		DRAW_PATTERN(lsr, 15,  6)
		DRAW_PATTERN(lsr, 19,  4)
		DRAW_PATTERN(lsr, 23,  2)
		DRAW_PATTERN(lsr, 27,  0)
		:
		: [p_buff] "l" (p_buff), [p_palette] "l" (p_palette), [pattern] "l" (pattern), [pmask] "l" (0x1e)
		: "r7"
	) ;
#else
	uint32_t	value ;


	value = (pattern <<  1) & 0x1e ;  if( value ) { p_buff[7] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >>  3) & 0x1e ;  if( value ) { p_buff[6] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >>  7) & 0x1e ;  if( value ) { p_buff[5] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 11) & 0x1e ;  if( value ) { p_buff[4] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 15) & 0x1e ;  if( value ) { p_buff[3] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 19) & 0x1e ;  if( value ) { p_buff[2] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 23) & 0x1e ;  if( value ) { p_buff[1] = *(uint16_t *)&p_palette[value] ; }
	value = (pattern >> 27) & 0x1e ;  if( value ) { p_buff[0] = *(uint16_t *)&p_palette[value] ; }
#endif//USEASM
}


#ifdef  VGA_CFG_BMP4
// -------------------------------------------------------------------------------- 
//!  １行分の 4bit ビットマップを描画バッファへ描画                                 
//!                                                                                 
//!  @param [out] p_buff   描画バッファアドレス                                     
//!  @param [in]  line     描画行                                                   
//!  @param [in]  p_layer  レイヤー                                                 
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawLineBmp4(uint16_t *p_buff, uint32_t line, VGALAYER *p_layer)
{
	uint8_t		const *p_palette ;
	uint16_t	*p_buffEnd ;
	uint32_t	const *p_data ;


	line += p_layer->y ;
	if( line >= VGA_IMAGE_HEIGHT ) {
		line -= VGA_IMAGE_HEIGHT ;
	}

	p_data    = (uint32_t const *)&p_layer->p_data[line*(VGA_IMAGE_WIDTH/2)] ;
	p_palette = &p_layer->p_palette[0] ;
	p_buff    = &p_buff[VGALINEBUFF_MARGIN] ;
	p_buffEnd = &p_buff[VGA_IMAGE_WIDTH] ;

	do {
		drawPattern(&p_buff[0], &p_palette[0], *p_data++) ;
		p_buff += 8 ;
	} while( p_buff != p_buffEnd ) ;
}
#endif//VGA_CFG_BMP4


#ifdef  VGA_CFG_BMP8
// -------------------------------------------------------------------------------- 
//!  １行分の 8bit ビットマップを描画バッファへ描画                                 
//!                                                                                 
//!  @param [out] p_buff   描画バッファアドレス                                     
//!  @param [in]  line     描画行                                                   
//!  @param [in]  p_layer  レイヤー                                                 
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawLineBmp8(uint16_t *p_buff, uint32_t line, VGALAYER *p_layer)
{
	uint8_t			value ;
	const uint8_t	*p_data ;
	uint16_t		*p_buffEnd ;
	const uint16_t	*p_palette ;


	line += p_layer->y ;
	if( line >= VGA_IMAGE_HEIGHT ) {
		line -= VGA_IMAGE_HEIGHT ;
	}

	p_palette = (const uint16_t *)&p_layer->p_palette[0] ;

	if( BIT_TEST(g_vgaInfo.bgMode, 1) ) {			// ストレッチあり
		uint8_t		valueL, valueR ;
		uint32_t	left, right, mask ;


		p_buff    = &p_buff[VGALINEBUFF_MARGIN] ;
		p_buffEnd = &p_buff[VGA_IMAGE_WIDTH] ;
		p_data    = &p_layer->p_data[line*VGA_BMPL_WIDTH] ;

		do {
			value = p_data[0] ;  if( value ) { p_buff[0] = p_palette[value] ; }
			value = p_data[3] ;  if( value ) { p_buff[4] = p_palette[value] ; }

			valueL = p_data[1] ;
			valueR = p_data[2] ;
			if( valueL | valueR ) {
				if( valueL ) {
					left      = p_palette[valueL] ;
					p_buff[1] = left ;
					right     = p_buff[2] ;

					if( valueR ) {
						right     = p_palette[valueR] ;
						p_buff[3] = right ;
					}
				} else {
					left      = p_buff[2] ;
					right     = p_palette[valueR] ;
					p_buff[3] = right ;
				}

				mask      = ~((left ^ right) & 0x0111) ;
				p_buff[2] = ((left & mask) + (right & mask)) / 2 ;
			}

			p_data += 4 ;
			p_buff += 5 ;
		} while( p_buff != p_buffEnd ) ;

	} else {										// ストレッチなし

		if( BIT_TEST(g_vgaInfo.bgMode, 0) ) {		// VGA_BGMODE_256x240
			p_buff    = &p_buff[VGALINEBUFF_MARGIN+(p_layer->x&63)] ;
			p_buffEnd = &p_buff[VGA_BMPL_WIDTH] ;
			p_data    = &p_layer->p_data[line*VGA_BMPL_WIDTH] ;
		} else {									// VGA_BGMODE_320x240
			p_buff    = &p_buff[VGALINEBUFF_MARGIN] ;
			p_buffEnd = &p_buff[VGA_BMPH_WIDTH] ;
			p_data    = &p_layer->p_data[line*VGA_BMPH_WIDTH] ;
		}

		do {
			value = p_data[0] ;  if( value ) { p_buff[0] = p_palette[value] ; }
			value = p_data[1] ;  if( value ) { p_buff[1] = p_palette[value] ; }
			value = p_data[2] ;  if( value ) { p_buff[2] = p_palette[value] ; }
			value = p_data[3] ;  if( value ) { p_buff[3] = p_palette[value] ; }
			value = p_data[4] ;  if( value ) { p_buff[4] = p_palette[value] ; }
			value = p_data[5] ;  if( value ) { p_buff[5] = p_palette[value] ; }
			value = p_data[6] ;  if( value ) { p_buff[6] = p_palette[value] ; }
			value = p_data[7] ;  if( value ) { p_buff[7] = p_palette[value] ; }

			p_data += 8 ;
			p_buff += 8 ;
		} while( p_buff != p_buffEnd ) ;

	}
}
#endif//VGA_CFG_BMP8


#ifdef  VGA_CFG_BMP16
// -------------------------------------------------------------------------------- 
//!  １行分の 16bit ビットマップを描画バッファへ描画                                
//!                                                                                 
//!  @param [out] p_buff   描画バッファアドレス                                     
//!  @param [in]  line     描画行                                                   
//!  @param [in]  p_layer  レイヤー                                                 
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawLineBmp16(uint16_t *p_buff, uint32_t line, VGALAYER *p_layer)
{
	uint16_t		value, *p_buffEnd ;
	const uint16_t	*p_data ;


	line += p_layer->y ;
	if( line >= VGA_IMAGE_HEIGHT ) {
		line -= VGA_IMAGE_HEIGHT ;
	}

	p_data    = (const uint16_t *)&p_layer->p_data[line*(VGA_IMAGE_WIDTH*2)] ;
	p_buff    = &p_buff[VGALINEBUFF_MARGIN] ;
	p_buffEnd = &p_buff[VGA_IMAGE_WIDTH] ;

	do {
		value = p_data[0] ;  if( !(value>>15) ) { p_buff[0] = value ; }
		value = p_data[1] ;  if( !(value>>15) ) { p_buff[1] = value ; }
		value = p_data[2] ;  if( !(value>>15) ) { p_buff[2] = value ; }
		value = p_data[3] ;  if( !(value>>15) ) { p_buff[3] = value ; }
		value = p_data[4] ;  if( !(value>>15) ) { p_buff[4] = value ; }
		value = p_data[5] ;  if( !(value>>15) ) { p_buff[5] = value ; }
		value = p_data[6] ;  if( !(value>>15) ) { p_buff[6] = value ; }
		value = p_data[7] ;  if( !(value>>15) ) { p_buff[7] = value ; }

		p_data += 8 ;
		p_buff += 8 ;
	} while( p_buff != p_buffEnd ) ;
}
#endif//VGA_CFG_BMP16


#if     defined(VGA_CFG_BGTILE) || defined(VGA_CFG_FGTILE)
// -------------------------------------------------------------------------------- 
//!  １行分のタイルを描画バッファへ描画                                             
//!                                                                                 
//!  @param [out] p_buff   描画バッファアドレス                                     
//!  @param [in]  line     描画行                                                   
//!  @param [in]  p_layer  レイヤー                                                 
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void VGAFUNC drawLineTile(uint16_t *p_buff, uint32_t line, VGALAYER *p_layer)
{
	const uint8_t	*p_palette ;																				// const uint8_t	*p_data, *p_palette ;
	uint16_t		*p_buffEnd ;
	uint32_t		y, yofs, attr, pattern ;																	// uint32_t			x, y, yofs, attr, pattern ;


	interp_set_config(interp1, 1, &m_config8x8) ;																//
	interp1->base[0] = (uintptr_t )p_layer->p_palette ;															//
	interp1->base[1] = (uintptr_t )p_layer->p_pattern ;															//

	if( p_layer->x & (VGA_TILE_IMAGE_WIDTH-1) ) {		// ｘ方向端数あり
		p_buff    = &p_buff[VGALINEBUFF_MARGIN-(p_layer->x&(VGA_TILE_IMAGE_WIDTH-1))] ;
		p_buffEnd = &p_buff[VGA_IMAGE_WIDTH+VGA_TILE_IMAGE_WIDTH] ;
	} else {											// ｘ方向端数なし
		p_buff    = &p_buff[VGALINEBUFF_MARGIN] ;
		p_buffEnd = &p_buff[VGA_IMAGE_WIDTH] ;
	}

	interp0->accum[0] = ((p_layer->x / VGA_TILE_IMAGE_WIDTH) & (VGA_TILE_WIDTH_COUNT - 1)) * 2 ;				// x      = ((p_layer->x / VGA_TILE_IMAGE_WIDTH) & (VGA_TILE_WIDTH_COUNT - 1)) * 2 ;
	interp0->accum[1] = (uintptr_t )p_buff ;																	// 
																												//
	y                 = (line + p_layer->y) & (VGA_TILE_HEIGHT - 1) ;											// y      = (line + p_layer->y) & (VGA_TILE_HEIGHT - 1) ;
	interp0->base[2]  = (uintptr_t )&p_layer->p_data[(y/VGA_TILE_IMAGE_HEIGHT)*VGA_TILE_WIDTH_COUNT*2] ;		// p_data = &p_layer->p_data[(y/VGA_TILE_IMAGE_HEIGHT)*VGA_TILE_WIDTH_COUNT*2] ;
	y                 = (y & (VGA_TILE_IMAGE_HEIGHT - 1)) * 4 ;													// y      = (y & (VGA_TILE_IMAGE_HEIGHT - 1)) * 4 ;

	do {
		attr              = (uint32_t )*(uint16_t *)interp0->peek[2] ;											// attr = *(uint16_t *)&p_data[x&(VGA_TILE_WIDTH_COUNT*2-1)] ;
		interp1->accum[0] = attr << 16 ;																		// x   += 2 ;

		// パレット
		p_palette = (const uint8_t *)interp1->peek[0] ;															// p_palette = &p_layer->p_palette[(attr>>VGATILE_PC_BIT)<<5] ;

		// 上下反転
		yofs = y ;
		if( BIT_TEST(attr, VGATILE_VF_BIT) ) {
			yofs ^= (7 * 4) ;
		}
		pattern = *(const uint32_t *)(interp1->peek[1] + yofs) ;												// pattern = *(uint32_t *)&p_layer->p_pattern[(attr<<(32-10)>>(32-10-5))|yofs] ;	// -5 = x32

		// 左右反転
		if( BIT_TEST(attr, VGATILE_HF_BIT) ) {
			drawPatternFlip(&p_buff[0], &p_palette[0], pattern) ;
		} else {
			drawPattern(&p_buff[0], &p_palette[0], pattern) ;
		}

		p_buff = (uint16_t *)interp0->pop[1] ;																	// p_buff += 8 ;
	} while( p_buff != p_buffEnd ) ;
}
#endif//defined(VGA_CFG_BGTILE) || defined(VGA_CFG_FGTILE)


#ifdef  VGA_CFG_SPR
// -------------------------------------------------------------------------------- 
//!  スプライト情報リストを作成                                                     
//!                                                                                 
//!  @param [in]  line            行                                                
//!  @param [out] pp_sprInfoList  スプライト情報リストの先頭アドレス                
//!  @param [out] p_sprInfo       スプライト情報アドレス                            
//!  @return                      なし                                              
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC createSprInfoList(uint32_t line, SPRINFO **pp_sprInfoList, SPRINFO *p_sprInfo)
{
	const VGASPR	*p_spr, *p_sprEnd ;
	SPRINFO			**pp_sprInfo, *p_sprInfoEnd ;
	uint32_t		y, attr ;


	p_spr        = (const VGASPR *)&g_vgaInfo.sprLayer.p_data[0] ;
	p_sprEnd     = &p_spr[VGA_SPR_MAX] ;
	p_sprInfoEnd = &p_sprInfo[VGA_SPR_LINE_MAX] ;

	pp_sprInfoList-- ;

	do {
		y = p_spr->y ;
		if( y >> VGASPR_Y_OFS_BIT ) {
			y += g_vgaInfo.sprLayer.y ;
		}

		y = (line - y) << (32-9) ;			// ｙ座標以外クリア
		if( !(y>>((32-9)+4)) ) {			// ｙ座標 x 4
			attr = p_spr->attr ;
			if( attr >> 30 ) {
				p_sprInfo->attr = attr ;

				// 上下反転
				y >>= (32-9) ;
				if( BIT_TEST(attr, VGASPR_ATTR_VF_BIT) ) {
					y ^= (VGA_SPR_IMAGE_HEIGHT-1) ;
				}

				p_sprInfo->y = y ;
				p_sprInfo->x = p_spr->x ;

				// リスト接続
				pp_sprInfo        = &pp_sprInfoList[attr>>VGASPR_ATTR_PR_BIT] ;
				p_sprInfo->p_next = *pp_sprInfo ;
				*pp_sprInfo       = p_sprInfo ;

				if( ++p_sprInfo == p_sprInfoEnd ) {
					break ;
				}
			}
		}
	} while( ++p_spr != p_sprEnd ) ;
}


// -------------------------------------------------------------------------------- 
//!  １行分のスプライトを描画バッファへ描画                                         
//!                                                                                 
//!  @param [out] p_buff     描画バッファアドレス                                   
//!  @param [in]  p_sprInfo  スプライト情報                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC VGAFUNC drawLineSpr(uint16_t *p_buff, SPRINFO *p_sprInfo)
{
	const uint8_t	*p_palette ;
	uint16_t		*p_dest ;
	int32_t			xofs ;
	uint32_t		x, attr ;
	const uint32_t	*p_pattern ;


	if( p_sprInfo == NULL ) {
		return ;
	}

	interp_set_config(interp1, 1, &m_config16x16) ;																//
	interp1->base[0] = (uintptr_t )g_vgaInfo.sprLayer.p_palette ;												//
	interp1->base[1] = (uintptr_t )g_vgaInfo.sprLayer.p_pattern ;												//

	xofs = g_vgaInfo.sprLayer.x ;

	do {
		x = p_sprInfo->x ;
		if( x >> VGASPR_X_OFS_BIT ) {
			x += xofs ;
		}

		x = (x + (VGA_SPR_IMAGE_WIDTH - 1)) & VGASPR_X_POS ;
		if( x < (VGA_IMAGE_WIDTH+VGA_SPR_IMAGE_WIDTH) ) {
			p_dest = &p_buff[x+1] ;

			attr              = (uint32_t )p_sprInfo->attr ;													// attr      = (uint16_t        )p_sprInfo->attr ;
			interp1->accum[0] = attr << 16 ;																	//
																												//
			p_palette = (const uint8_t  *)interp1->peek[0] ;													// p_palette = (const uint8_t  *)&g_vgaInfo.sprLayer.p_palette[(attr>>VGASPR_ATTR_PC_BIT)<<5] ;
			p_pattern = (const uint32_t *)(interp1->peek[1] + (p_sprInfo->y * 4)) ;								// p_pattern = (const uint32_t *)&g_vgaInfo.sprLayer.p_pattern[((attr<<(32-9))>>(32-9-7))+(p_sprInfo->y*4)] ;		// -7 = x128

			// 左右反転
			if( BIT_TEST(attr, VGASPR_ATTR_HF_BIT) ) {
				drawPatternFlip(&p_dest[0], &p_palette[0], p_pattern[16]) ;
				drawPatternFlip(&p_dest[8], &p_palette[0], p_pattern[0]) ;
			} else {
				drawPattern(&p_dest[0], &p_palette[0], p_pattern[0]) ;
				drawPattern(&p_dest[8], &p_palette[0], p_pattern[16]) ;
			}
		}

		p_sprInfo = p_sprInfo->p_next ;
	} while( p_sprInfo != NULL ) ;
}
#endif//VGA_CFG_SPR


#ifdef  VGA_CFG_TXT
// -------------------------------------------------------------------------------- 
//!  １行分のテキストを描画バッファへ描画                                           
//!                                                                                 
//!  @param [out] p_buff   描画バッファアドレス                                     
//!  @param [in]  line     描画行                                                   
//!  @param [in]  p_layer  レイヤー                                                 
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static inline void VGAFUNC drawLineTxt(uint16_t *p_buff, uint32_t line, VGALAYER *p_layer)
{
	const uint8_t		*p_palette ;
	uint16_t			*p_buffEnd ;
	uint32_t			attr, pattern ;

#ifndef USEASM
	uint32_t			value ;
#endif//USEASM

	line += p_layer->y ;
	if( line >= VGA_IMAGE_HEIGHT ) {
		line -= VGA_IMAGE_HEIGHT ;
	}

	interp_set_config(interp1, 1, &m_config8x8) ;																		//
	interp1->base[0]  = (uintptr_t )p_layer->p_palette ;																//
	interp1->base[1]  = (uintptr_t )&p_layer->p_pattern[(line&(VGA_TXT_IMAGE_HEIGHT-1))*4] ;							// p_pattern = &p_layer->p_pattern[(line&(VGA_TXT_IMAGE_HEIGHT-1))*4] ;

	interp0->accum[0] = 0 ;																								//
	interp0->accum[1] = (uintptr_t )p_buff ;																			//
	interp0->base[2]  = (uintptr_t )&p_layer->p_data[(line/VGA_TXT_IMAGE_HEIGHT)*(VGA_TXT_WIDTH_COUNT*2)] ;				// p_txt = (const uint16_t *)&p_layer->p_data[(line/VGA_TXT_IMAGE_HEIGHT)*(VGA_TXT_WIDTH_COUNT*2)] ;
																														//
	p_buffEnd = &p_buff[VGA_IMAGE_WIDTH] ;

	do {
		attr              = (uint32_t )*(uint16_t *)interp0->peek[2] ;													// attr      = *p_txt++ ;
		interp1->accum[0] = attr << 16 ;																				//
																														//
		p_palette =  (const uint8_t  *)interp1->peek[0] ;																// p_palette = &p_layer->p_palette[(attr>>VGATXT_PC_BIT)<<5] ;
		pattern   = *(const uint32_t *)interp1->peek[1] ;																// pattern   = *(uint32_t *)&p_pattern[(attr<<(32-9))>>(32-9-5)] ;				// -5 = x32

		if( BIT_TEST(attr, VGATXT_SH_BIT) ) {		// シャドーあり

			if( BIT_TEST(attr, VGATXT_OP_BIT) ) {	// 透過なし
#ifdef  USEASM
				asm volatile (
					DRAW_TXT_SH_OP(lsl,  1,  0)
					DRAW_TXT_SH_OP(lsr,  3,  2)
					DRAW_TXT_SH_OP(lsr,  7,  4)
					DRAW_TXT_SH_OP(lsr, 11,  6)
					DRAW_TXT_SH_OP(lsr, 15,  8)
					DRAW_TXT_SH_OP(lsr, 19, 10)
					DRAW_TXT_SH_OP(lsr, 23, 12)
					DRAW_TXT_SH_OP(lsr, 27, 14)
					:
					: [p_buff] "l" (p_buff), [p_palette] "l" (p_palette), [pattern] "l" (pattern), [pmask] "l" (0x1e), [smask] "l" (0x0eee)
					: "r7"
				) ;
#else
				value = (pattern <<  1) & 0x1e ;  if( value ) { p_buff[0] = *(uint16_t *)&p_palette[value] ; } else { p_buff[0] = (p_buff[0] & 0x0eee) / 2 ; }
				value = (pattern >>  3) & 0x1e ;  if( value ) { p_buff[1] = *(uint16_t *)&p_palette[value] ; } else { p_buff[1] = (p_buff[1] & 0x0eee) / 2 ; }
				value = (pattern >>  7) & 0x1e ;  if( value ) { p_buff[2] = *(uint16_t *)&p_palette[value] ; } else { p_buff[2] = (p_buff[2] & 0x0eee) / 2 ; }
				value = (pattern >> 11) & 0x1e ;  if( value ) { p_buff[3] = *(uint16_t *)&p_palette[value] ; } else { p_buff[3] = (p_buff[3] & 0x0eee) / 2 ; }
				value = (pattern >> 15) & 0x1e ;  if( value ) { p_buff[4] = *(uint16_t *)&p_palette[value] ; } else { p_buff[4] = (p_buff[4] & 0x0eee) / 2 ; }
				value = (pattern >> 19) & 0x1e ;  if( value ) { p_buff[5] = *(uint16_t *)&p_palette[value] ; } else { p_buff[5] = (p_buff[5] & 0x0eee) / 2 ; }
				value = (pattern >> 23) & 0x1e ;  if( value ) { p_buff[6] = *(uint16_t *)&p_palette[value] ; } else { p_buff[6] = (p_buff[6] & 0x0eee) / 2 ; }
				value = (pattern >> 27) & 0x1e ;  if( value ) { p_buff[7] = *(uint16_t *)&p_palette[value] ; } else { p_buff[7] = (p_buff[7] & 0x0eee) / 2 ; }
#endif//USEASM
			} else {								// 透過あり
#ifdef  USEASM
				asm volatile (
					DRAW_TXT_SH(lsl,  1,  0)
					DRAW_TXT_SH(lsr,  3,  2)
					DRAW_TXT_SH(lsr,  7,  4)
					DRAW_TXT_SH(lsr, 11,  6)
					DRAW_TXT_SH(lsr, 15,  8)
					DRAW_TXT_SH(lsr, 19, 10)
					DRAW_TXT_SH(lsr, 23, 12)
					DRAW_TXT_SH(lsr, 27, 14)
					:
					: [p_buff] "l" (p_buff), [pattern] "l" (pattern), [pmask] "l" (0x1e), [smask] "l" (0x0eee)
					: "r7"
				) ;
#else
				if(  pattern      << 28 ) { p_buff[0] = (p_buff[0] & 0x0eee) / 2 ; }
				if( (pattern>> 4) << 28 ) { p_buff[1] = (p_buff[1] & 0x0eee) / 2 ; }
				if( (pattern>> 8) << 28 ) { p_buff[2] = (p_buff[2] & 0x0eee) / 2 ; }
				if( (pattern>>12) << 28 ) { p_buff[3] = (p_buff[3] & 0x0eee) / 2 ; }
				if( (pattern>>16) << 28 ) { p_buff[4] = (p_buff[4] & 0x0eee) / 2 ; }
				if( (pattern>>20) << 28 ) { p_buff[5] = (p_buff[5] & 0x0eee) / 2 ; }
				if( (pattern>>24) << 28 ) { p_buff[6] = (p_buff[6] & 0x0eee) / 2 ; }
				if( (pattern>>28)       ) { p_buff[7] = (p_buff[7] & 0x0eee) / 2 ; }
#endif//USEASM
			}

		} else {									// シャドーなし

			if( BIT_TEST(attr, VGATXT_OP_BIT) ) {	// 透過なし
#ifdef  USEASM
				asm volatile (
					DRAW_TXT_OP(lsl,  1,  0)
					DRAW_TXT_OP(lsr,  3,  2)
					DRAW_TXT_OP(lsr,  7,  4)
					DRAW_TXT_OP(lsr, 11,  6)
					DRAW_TXT_OP(lsr, 15,  8)
					DRAW_TXT_OP(lsr, 19, 10)
					DRAW_TXT_OP(lsr, 23, 12)
					DRAW_TXT_OP(lsr, 27, 14)
					:
					: [p_buff] "l" (p_buff), [p_palette] "l" (p_palette), [pattern] "l" (pattern), [pmask] "l" (0x1e)
					: "r7"
				) ;
#else
				p_buff[0] = *(uint16_t *)&p_palette[(pattern<< 1)&0x1e] ;
				p_buff[1] = *(uint16_t *)&p_palette[(pattern>> 3)&0x1e] ;
				p_buff[2] = *(uint16_t *)&p_palette[(pattern>> 7)&0x1e] ;
				p_buff[3] = *(uint16_t *)&p_palette[(pattern>>11)&0x1e] ;
				p_buff[4] = *(uint16_t *)&p_palette[(pattern>>15)&0x1e] ;
				p_buff[5] = *(uint16_t *)&p_palette[(pattern>>19)&0x1e] ;
				p_buff[6] = *(uint16_t *)&p_palette[(pattern>>23)&0x1e] ;
				p_buff[7] = *(uint16_t *)&p_palette[(pattern>>27)&0x1e] ;
#endif//USEASM
			} else {								// 透過あり
#ifdef  USEASM
				asm volatile (
					DRAW_PATTERN(lsl,  1,  0)
					DRAW_PATTERN(lsr,  3,  2)
					DRAW_PATTERN(lsr,  7,  4)
					DRAW_PATTERN(lsr, 11,  6)
					DRAW_PATTERN(lsr, 15,  8)
					DRAW_PATTERN(lsr, 19, 10)
					DRAW_PATTERN(lsr, 23, 12)
					DRAW_PATTERN(lsr, 27, 14)
					:
					: [p_buff] "l" (p_buff), [p_palette] "l" (p_palette), [pattern] "l" (pattern), [pmask] "l" (0x1e)
					: "r7"
				) ;
#else
				drawPattern(&p_buff[0], &p_palette[0], pattern) ;
#endif//USEASM
			}

		}

		p_buff = (uint16_t *)interp0->pop[1] ;																	// p_buff += 8 ;
	} while( p_buff != p_buffEnd ) ;
}
#endif//VGA_CFG_TXT


// -------------------------------------------------------------------------------- 
//!  １行描画                                                                       
//!                                                                                 
//!  @param [out] p_buff  描画バッファアドレス                                      
//!  @param [in]  line    描画行                                                    
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static void VGAFUNC drawLine(uint16_t *p_buff, uint32_t line)
{
#ifdef  VGA_CFG_SPR
	SPRINFO		sprInfo[VGA_SPR_LINE_MAX], *p_sprInfoList[3] ;
#endif//VGA_CFG_SPR

	// HSync 割り込み
	if( g_vgaInfo.hsyncProc != NULL ) {
		g_vgaInfo.hsyncProc( line ) ;
	}

	// 背景塗り潰し開始
	p_buff[0] = g_vgaInfo.color ;
	p_buff[1] = g_vgaInfo.color ;
	channel_config_set_read_increment(&g_vgaInfo.dmaConfigData, false) ;
	dma_channel_configure(VGA_CFG_DMACH_DATA, &g_vgaInfo.dmaConfigData, &p_buff[VGALINEBUFF_MARGIN], &p_buff[0], (VGA_IMAGE_WIDTH*2)/4, true) ;

#ifdef  VGA_CFG_SPR
	p_sprInfoList[0] = NULL ;
	p_sprInfoList[1] = NULL ;
	p_sprInfoList[2] = NULL ;

	// スプライト情報作成
	if( g_vgaInfo.sprLayer.p_data != NULL ) {
		createSprInfoList(line, &p_sprInfoList[0], &sprInfo[0]) ;
	}
#endif//VGA_CFG_SPR

	// 背景塗り潰し完了待ち
	dma_channel_wait_for_finish_blocking( VGA_CFG_DMACH_DATA ) ;

	// ---------------------------------------------------------------- 
	//   奥スプライト描画                                               
	// ---------------------------------------------------------------- 
#ifdef  VGA_CFG_SPR
	if( g_vgaInfo.sprLayer.p_data != NULL ) {
		drawLineSpr(&p_buff[0], p_sprInfoList[0]) ;
	}
#endif//VGA_CFG_SPR

	// ---------------------------------------------------------------- 
	//   背景描画                                                       
	// ---------------------------------------------------------------- 
#if     defined(VGA_CFG_BMP4) || defined(VGA_CFG_BMP8) || defined(VGA_CFG_BMP16) || defined(VGA_CFG_BGTILE)
	switch( g_vgaInfo.bgMode >> 4 ) {
#ifdef  VGA_CFG_BMP4
		case (VGA_BGMODE_BMP4>>4):  { drawLineBmp4(&p_buff[0],  line, &g_vgaInfo.bgLayer) ;  break ; }
#endif//VGA_CFG_BMP4

#ifdef  VGA_CFG_BMP8
		case (VGA_BGMODE_BMP8>>4):  { drawLineBmp8(&p_buff[0],  line, &g_vgaInfo.bgLayer) ;  break ; }
#endif//VGA_CFG_BMP8

#ifdef  VGA_CFG_BMP16
		case (VGA_BGMODE_BMP16>>4): { drawLineBmp16(&p_buff[0], line, &g_vgaInfo.bgLayer) ;  break ; }
#endif//VGA_CFG_BMP16

#ifdef  VGA_CFG_BGTILE
		case (VGA_BGMODE_TILE>>4):  { drawLineTile(&p_buff[0],  line, &g_vgaInfo.bgLayer) ;  break ; }
#endif//VGA_CFG_BGTILE
//		case VGA_BGMODE_BLANK:
		default:
			break ;
	}
#endif//defined(VGA_CFG_BMP4) || defined(VGA_CFG_BMP8) || defined(VGA_CFG_BMP16) || defined(VGA_CFG_BGTILE)

	// ---------------------------------------------------------------- 
	//   中スプライト描画                                               
	// ---------------------------------------------------------------- 
#ifdef  VGA_CFG_SPR
	if( g_vgaInfo.sprLayer.p_data != NULL ) {
		drawLineSpr(&p_buff[0], p_sprInfoList[1]) ;
	}
#endif//VGA_CFG_SPR

	// ---------------------------------------------------------------- 
	//   前景描画                                                       
	// ---------------------------------------------------------------- 
#ifdef  VGA_CFG_FGTILE
	if( g_vgaInfo.fgLayer.p_data != NULL ) {
		drawLineTile(&p_buff[0], line, &g_vgaInfo.fgLayer) ;
	}

	// ---------------------------------------------------------------- 
	//   手前スプライト描画                                             
	// ---------------------------------------------------------------- 
#ifdef  VGA_CFG_SPR
	if( g_vgaInfo.sprLayer.p_data != NULL ) {
		drawLineSpr(&p_buff[0], p_sprInfoList[2]) ;
	}
#endif//VGA_CFG_SPR
#endif//VGA_CFG_FGTILE

	// ---------------------------------------------------------------- 
	//   テキスト描画                                                   
	// ---------------------------------------------------------------- 
#ifdef  VGA_CFG_TXT
	if( g_vgaInfo.txtLayer.p_data != NULL ) {
		drawLineTxt(&p_buff[VGALINEBUFF_MARGIN], line, &g_vgaInfo.txtLayer) ;
	}
#endif//VGA_CFG_TXT
}


// -------------------------------------------------------------------------------- 
//!  メイン処理                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC mainProc( void )
{
	uint32_t	line ;


#ifdef  DEBUG_PERFORMANCE
	gpio_init( DEBUG_PERFORMANCE ) ;
	gpio_set_dir(DEBUG_PERFORMANCE, GPIO_OUT) ;
	gpio_put(DEBUG_PERFORMANCE, 1) ;
#endif//DEBUG_PERFORMANCE

	init() ;

	while( multicore_fifo_rvalid() == false ) {

		line = g_vgaInfo.drawLine ;
		if( line ) {

			// 先頭行
			if( line == 1 ) {
				drawLine(&g_vgaInfo.lineBuff.page[0].margin[0], 0) ;
				dma_channel_configure(VGA_CFG_DMACH_PIXEL, &g_vgaInfo.dmaConfigPixel, &PIO->txf[SM_TRANS_PIXEL], &g_vgaInfo.lineBuff.page[0].data[0], (VGA_IMAGE_WIDTH*2)/sizeof( uint32_t ), true) ;
			}

#ifdef  DEBUG_PERFORMANCE
			gpio_put(DEBUG_PERFORMANCE, 0) ;
#endif//DEBUG_PERFORMANCE

			drawLine(&g_vgaInfo.lineBuff.page[line&1].margin[0], line) ;

#ifdef  DEBUG_PERFORMANCE
			gpio_put(DEBUG_PERFORMANCE, 1) ;
#endif//DEBUG_PERFORMANCE

			g_vgaInfo.drawLine = 0 ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void vga_init( void )
{
	memset(&g_vgaInfo, 0, sizeof( g_vgaInfo )) ;
}


// -------------------------------------------------------------------------------- 
//!  開始                                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void vga_start( void )
{
	multicore_launch_core1( mainProc ) ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void VGAFUNC vga_waitVSync( void )
{
	while( !gpio_get( VGA_CFG_PIN_VSYNC ) ) {
		;
	}

	while( gpio_get( VGA_CFG_PIN_VSYNC ) ) {
		;
	}
}
