// -------------------------------------------------------------------------------- 
//!  @file  menu.c                                                                  
//!  メニューのソースファイル                                                       
//                                                                                  
//   Copyright (C) 2022 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"

typedef struct _MENUITEM			MENUITEM ;


	enum {
		MENUITEM_X					=	2,							//!< メニュー項目描画ｘ座標
		PARAM_X						=	16							//!< パラメータ描画ｘ座標
	} ;

	//! メニュー項目
	struct _MENUITEM {
		uint8_t						x ;								//!< ｘ座標
		uint8_t						y ;								//!< ｙ座標
		char						const *p_string ;				//!< 文字列
		void						(*proc)(MENUITEM *p_menuItem, uint16_t pad) ;
																	//!< 処理
		void						(*draw)(MENUITEM *p_menuItem, uint palette) ;
																	//!< 描画処理
	} ;


static char							m_offOn[][3]={					//!< オフ・オン
										"OFF",
										"ON "
									},
									m_language[][5]={				//!< 言語
										"JAPAN",
										"OTHER"
									},
									m_button[][4]={					//!< ボタン名
										"NONE",
										"    ",
										"BTN2",
										"BTN3",
										"BTN4",
										"BTN5",
										"BTN6",
										"BTN7",
										"LB  ",
										"RB  ",
									},
									m_padType[][14]={				//!< パッド種類
										"NORMAL        ",
										"PADDLE CONTROL"
									} ;


// -------------------------------------------------------------------------------- 
//!  ゲームモードへ切り替える                                                       
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC gameMode( void )
{
	g.menu = 0 ;
	memset(&g.txt[0][0], 0, sizeof( g.txt )) ;

	sound_mute( 0 ) ;
}


// -------------------------------------------------------------------------------- 
//!  文字列を書き込む                                                               
//!                                                                                 
//!  @param [in] x         ｘ座標                                                   
//!  @param [in] y         ｙ座標                                                   
//!  @param [in] p_string  文字列                                                   
//!  @param [in] length    最大文字数                                               
//!  @param [in] palette   パレット番号                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawString(int x, int y, char const *p_string, int length, uint palette)
{
	uint8_t		code ;
	uint16_t	*p_txt ;


	p_txt     = &g.txt[y][x] ;
	palette <<= VGATXT_PC_BIT ;

	while( length-- ) {
		code = (uint8_t )*p_string++ ;
		if( code == '\0' ) {
			break ;
		}

		*p_txt++ = code | palette ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ゲームメニュー処理                                                             
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC game(MENUITEM *p_menuItem, uint16_t pad)
{
	if( g.pad.once & PAD_BTN(1) ) {
		gameMode() ;
		sm3gg_setRomInfo( &g_romInfo[g.select] ) ;
		sound_reset() ;
		return ;
	}

	if( g.pad.once & (PAD_LEFT|PAD_RIGHT) ) {
		if( g.pad.once & PAD_LEFT ) {
			if( --g.select < 0 ) {
				g.select = (sizeof( g_romInfo ) / sizeof( g_romInfo[0] )) - 1 ;
			}
		} else {
			if( ++g.select == (sizeof( g_romInfo )/sizeof( g_romInfo[0] )) ) {
				g.select = 0 ;
			}
		}

		drawString(PARAM_X, 4, &g_romInfo[g.select].name[0], sizeof( g_romInfo[0].name ), 1) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ゲームメニュー描画処理                                                         
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawGame(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, 4, &g_romInfo[g.select].name[0], sizeof( g_romInfo[0].name ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  リセットメニュー処理                                                           
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC reset(MENUITEM *p_menuItem, uint16_t pad)
{
	if( pad & PAD_BTN(1) ) {
		gameMode() ;
		sm3gg_reset() ;
		sound_reset() ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ワイドメニュー処理                                                             
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC wideScreen(MENUITEM *p_menuItem, uint16_t pad)
{
	if( pad & (PAD_LEFT|PAD_RIGHT|PAD_BTN(1)) ) {
		g.wide ^= 1 ;
		drawString(PARAM_X, p_menuItem->y, &m_offOn[g.wide][0], sizeof( m_offOn[0] ), 1) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ワイドメニュー描画処理                                                         
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawWideScreen(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_offOn[g.wide][0], sizeof( m_offOn[0] ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  言語メニュー処理                                                               
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC language(MENUITEM *p_menuItem, uint16_t pad)
{
	if( pad & (PAD_LEFT|PAD_RIGHT|PAD_BTN(1)) ) {
		g.language ^= 1 ;
		drawString(PARAM_X, p_menuItem->y, &m_language[g.language][0], sizeof( m_language[0] ), 1) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  言語メニュー描画処理                                                           
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawLanguage(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_language[g.language][0], sizeof( m_language[0] ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  ラピッドファイアメニュー処理                                                   
//!                                                                                 
//!  @param [in] p_menuItem   メニュー項目                                          
//!  @param [in] pad          パッド状態                                            
//!  @param [in] p_rapidFire  ラピッドファイア                                      
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
static void RAMFUNC rapidFire(MENUITEM *p_menuItem, uint16_t pad, RAPIDFIRE *p_rapidFire)
{
	int		button ;


	button = 0 ;

	if( !(pad&PAD_BTN(1)) ) {
		for(button=PAD_BTN_+2; button<=PAD_RB_; button++) {
			if( pad & (1<<button) ) {
				break ;
			}
		}
		if( button > PAD_RB_ ) {
			return ;
		}
	}

	if( button == p_rapidFire->button ) {
		return ;
	}

	p_rapidFire->button = button ;

	drawString(PARAM_X, p_menuItem->y, &m_button[p_rapidFire->button][0], sizeof( m_button[0] ), 1) ;
}


// -------------------------------------------------------------------------------- 
//!  ラピッドファイア１メニュー処理                                                 
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC rapidFire1(MENUITEM *p_menuItem, uint16_t pad)
{
	rapidFire(p_menuItem, pad, &g.rapidFire[0]) ;
}


// -------------------------------------------------------------------------------- 
//!  ラピッドファイア１メニュー描画処理                                             
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawRapidFire1(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_button[g.rapidFire[0].button][0], sizeof( m_button[0] ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  ラピッドファイア２メニュー処理                                                 
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC rapidFire2(MENUITEM *p_menuItem, uint16_t pad)
{
	rapidFire(p_menuItem, pad, &g.rapidFire[1]) ;
}


// -------------------------------------------------------------------------------- 
//!  ラピッドファイア２メニュー描画処理                                             
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawRapidFire2(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_button[g.rapidFire[1].button][0], sizeof( m_button[0] ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  パッド種類メニュー処理                                                         
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC padType(MENUITEM *p_menuItem, uint16_t pad)
{
	if( pad & (PAD_LEFT|PAD_RIGHT|PAD_BTN(1)) ) {
		g.pad.type ^= 1 ;
		drawString(PARAM_X, p_menuItem->y, &m_padType[g.pad.type][0], sizeof( m_padType[0] ), 1) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  パッド種類メニュー描画処理                                                     
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawPadType(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_padType[g.pad.type][0], sizeof( m_padType[0] ), palette) ;
}


// -------------------------------------------------------------------------------- 
//!  ボーダー色メニュー処理                                                         
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] pad         パッド状態                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC borderColor(MENUITEM *p_menuItem, uint16_t pad)
{
	if( pad & (PAD_LEFT|PAD_RIGHT|PAD_BTN(1)) ) {
		g.borderColor ^= 1 ;
		drawString(PARAM_X, p_menuItem->y, &m_offOn[g.borderColor][0], sizeof( m_offOn[0] ), 1) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ボーダー色メニュー描画処理                                                     
//!                                                                                 
//!  @param [in] p_menuItem  メニュー項目                                           
//!  @param [in] palette     パレット番号                                           
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawBorderColor(MENUITEM *p_menuItem, uint palette)
{
	drawString(PARAM_X, p_menuItem->y, &m_offOn[g.borderColor][0], sizeof( m_offOn[0] ), palette) ;
}

//! メニュー項目
static MENUITEM						m_menuItem[]={
										{ MENUITEM_X,  4, "GAME",         game,        drawGame        },
										{ MENUITEM_X,  5, "RESET",        reset,       NULL            },
										{ MENUITEM_X,  6, "WIDE SCREEN",  wideScreen,  drawWideScreen  },
										{ MENUITEM_X,  7, "LANGUAGE",     language,    drawLanguage    },
										{ MENUITEM_X,  8, "RAPID FIRE 1", rapidFire1,  drawRapidFire1  },
										{ MENUITEM_X,  9, "RAPID FIRE 2", rapidFire2,  drawRapidFire2  },
										{ MENUITEM_X, 10, "PAD TYPE",     padType,     drawPadType     },
										{ MENUITEM_X, 11, "BORDER COLOR", borderColor, drawBorderColor },
									} ;


// -------------------------------------------------------------------------------- 
//!  メニュー処理                                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC menuProc( void )
{
	MENUITEM	*p_menuItem ;
	int			index ;

	// メニュー終了
	if( g.pad.once & (PAD_SELECT|PAD_BTN(0)) ) {
		g.menu = 0 ;
		memset(&g.txt[0][0], 0, sizeof( g.txt )) ;
		sound_mute( 0 ) ;
		return ;
	}

	switch( g.proc ) {
		case 0:
		{
			uint16_t	*p_palette, *p_paletteEnd ;

			// パレット輝度下げる
			g.p_vram->borderColor = (g.p_vram->borderColor & 0x0eee0eee) / 2 ;
			vga_setColor( g.p_vram->borderColor ) ;

			p_palette    = &g.p_vram->palette[0][0] ;
			p_paletteEnd = &g.p_vram->palette[192][0] ;

			do {
				*(uint32_t *)&p_palette[0] = (*(uint32_t *)&p_palette[0] & 0x0eee0eee) / 2 ;
				*(uint32_t *)&p_palette[2] = (*(uint32_t *)&p_palette[2] & 0x0eee0eee) / 2 ;
				*(uint32_t *)&p_palette[4] = (*(uint32_t *)&p_palette[4] & 0x0eee0eee) / 2 ;
				*(uint32_t *)&p_palette[6] = (*(uint32_t *)&p_palette[6] & 0x0eee0eee) / 2 ;
				p_palette                 += 8 ;
			} while( p_palette != p_paletteEnd ) ;

			// メニュー描画
			p_menuItem = &m_menuItem[0] ;
			for(index=0; index<sizeof( m_menuItem )/sizeof( m_menuItem[0] ); index++,p_menuItem++) {
				drawString(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0],  -1, index==g.cursor) ;

				if( p_menuItem->draw != NULL ) {
					p_menuItem->draw(p_menuItem, index==g.cursor) ;
				}
			}
			g.proc++ ;
			break ;
		}

		case 1:
		{
			// カーソル移動
			if( g.pad.once & (PAD_UP|PAD_DOWN) ) {
				p_menuItem = &m_menuItem[g.cursor] ;
				drawString(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], -1, 0) ;

				if( p_menuItem->draw != NULL ) {
					p_menuItem->draw(p_menuItem, 0) ;
				}

				if( g.pad.once & PAD_UP ) {
					if( --g.cursor < 0 ) {
						g.cursor = sizeof( m_menuItem ) / sizeof( m_menuItem[0] ) - 1 ;
					}
				} else {
					if( ++g.cursor == (sizeof( m_menuItem )/sizeof( m_menuItem[0] )) ) {
						g.cursor = 0 ;
					}
				}

				p_menuItem = &m_menuItem[g.cursor] ;
				drawString(p_menuItem->x, p_menuItem->y, &p_menuItem->p_string[0], -1, 1) ;

				if( p_menuItem->draw != NULL ) {
					p_menuItem->draw(p_menuItem, 1) ;
				}
				break ;
			}

			m_menuItem[g.cursor].proc(&m_menuItem[g.cursor], g.pad.once) ;
			break ;
		}
	}
}
