// -------------------------------------------------------------------------------- 
//!  @file  sm3gg.c                                                                 
//!  SEGA MARKIII / GAME GEAR のソースファイル                                      
//                                                                                  
//   Copyright (C) 2022 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"

	//! メモリ情報
	typedef struct {
		uint8_t						const *p_read ;					//!< 読み込み領域
		uint8_t						*p_write ;						//!< 書き込み領域
	} MEMINFO ;

	//! SEGA MARK III / GAME GEAR
	typedef struct {
		uint8_t						const *p_rom ;					//!< ROM
		uint32_t					romMask ;						//!< ROM マスク
		uint16_t					ramBase[4] ;					//!< RAM ベースアドレス
		uint8_t						bank[4] ;						//!< バンク
		MEMINFO						memInfo[8] ;					//!< メモリ情報
		ADDRESSBUS					addressBusIo ;					//!< I/O アドレスバス
		ADDRESSBUS					addressBusMem ;					//!< メモリーアドレスバス
		Z80							z80 ;							//!< Z80
		VDP							vdp ;							//!< VDP
		uint8_t						ram[8192] ;						//!< RAM
		uint8_t						extRam[32768] ;					//!< 拡張 RAM

		// GAME GEAR
		struct {
			uint8_t					extData ;						//!< 拡張コネクタデータ
			uint8_t					extSet ;						//!< 拡張コネクタ設定
			uint8_t					sendData ;						//!< 送信データ
			uint8_t					comMode ;						//!< 通信モード
		} gg ;
	} SM3GG ;


static SM3GG						m_sm3gg ;						//!< SEGA MARK III / GAME GEAR


// -------------------------------------------------------------------------------- 
//!  ROM バンク設定                                                                 
//!                                                                                 
//!  @param [in] address    アドレス                                                
//!  @param [in] romAddress  ROM アドレス                                           
// -------------------------------------------------------------------------------- 
static void RAMFUNC setRomBank(uint32_t address, uint32_t romAddress)
{
	MEMINFO		*p_memInfo ;


	p_memInfo            = &m_sm3gg.memInfo[address/0x2000] ;
	p_memInfo[0].p_read  = &m_sm3gg.p_rom[(romAddress+0x0000)&m_sm3gg.romMask] ;
	p_memInfo[0].p_write = NULL ;
	p_memInfo[1].p_read  = &m_sm3gg.p_rom[(romAddress+0x2000)&m_sm3gg.romMask] ;
	p_memInfo[1].p_write = NULL ;
}


// -------------------------------------------------------------------------------- 
//!  RAM バンク設定                                                                 
//!                                                                                 
//!  @param [in] address    アドレス                                                
//!  @param [in] ramAddress  RAM アドレス                                           
// -------------------------------------------------------------------------------- 
static void RAMFUNC setRamBank(uint32_t address, uint32_t ramAddress)
{
	MEMINFO		*p_memInfo ;
	uint16_t	*p_ramBase ;


	p_ramBase = &m_sm3gg.ramBase[(ramAddress&(sizeof( m_sm3gg.extRam )-1))/0x2000] ;

	p_memInfo            = &m_sm3gg.memInfo[address/0x2000] ;
	p_memInfo[0].p_read  = p_memInfo[0].p_write = &m_sm3gg.extRam[p_ramBase[0]] ;
	p_memInfo[1].p_read  = p_memInfo[1].p_write = &m_sm3gg.extRam[p_ramBase[1]] ;
}


// -------------------------------------------------------------------------------- 
//!  アドレスバス読み取り                                                           
//!                                                                                 
//!  @param [in,out] p_this   this ポインタ                                         
//!  @param          address  アドレス                                              
//!  @return         データ                                                         
// -------------------------------------------------------------------------------- 
static uint8_t RAMFUNC addressbus_read_io(ADDRESSBUS *p_this, uint16_t address)
{
	uint8_t		data ;


	switch( (address/0x40) & 3 ) {
		case 0:		// 0x00 - 0x3f
		{
			if( g.mode == MODE_SM3 ) {
				data = 0xff ;
				break ;
			}

			switch( address & 0x3f ) {
				case 0x00:
				{
					// |STT|NJAP|NNTS|*|*|*|*|*|
					//
					// STT   START/PAUSE 0:on, 1:off
					// NJAP  0:Japan mode, 1:overseas mode
					// NNTS  0:NTSC,       1:PAL mode

					data = (((~g.pad.norm>>PAD_START_)&1)<<7) | (g.language<<6) ;
					break ;
				}

				case 0x01:
				{
					data = ((m_sm3gg.gg.extData & ~m_sm3gg.gg.extSet) & 0x7f) ;
					break ;
				}

				case 0x02:
				{
					data = m_sm3gg.gg.extSet ;
					break ;
				}

				case 0x03:
				{
					data = m_sm3gg.gg.sendData ;
					break ;
				}

				case 0x05:
				{
					data = m_sm3gg.gg.comMode ;
					break ;
				}

//				case 0x04:
//				case 0x30:
//				case 0x31:
				default:
				{
					data = 0xff ;
					break ;
				}
			}
			break ;
		}

		case 1:		// 0x40 - 0x7f
		{
			data = readAddressBus(&m_sm3gg.vdp.addressBus, (address&1)+2) ;
			break ;
		}

		case 2:		// 0x80 - 0xbf
		{
			data = readAddressBus(&m_sm3gg.vdp.addressBus, address&1) ;
			break ;
		}

		default:	// 0xc0 - 0xff
		{
			int		analog ;


			if( address & 1 ) {
				// |‐|‐|‐|‐|2P:２|2P:１|2P:右|2P:左|
				data = 0xff ;
				break ;
			}

			if( g.pad.type == PAD_TYPE_NORMAL ) {	// ジョイパッド

				// |2P:下|2P:上|1P/CK:２|1P/CK:１|1P/CK:右|1P/CK:左|1P/CK:下|1P/CK:上|
				data = ~((((g.pad.norm>>PAD_UP_)   &1)<<0) |
				         (((g.pad.norm>>PAD_DOWN_) &1)<<1) |
				         (((g.pad.norm>>PAD_LEFT_) &1)<<2) |
				         (((g.pad.norm>>PAD_RIGHT_)&1)<<3) |
				         (((g.pad.norm>>PAD_BTN_)  &3)<<4)) ;

			} else {								// パドルコントロール

				if( g.pad.paddle.pdsel < 8 ) {
					if( !g.pad.paddle.pdsel ) {
						analog = (int )g.pad.analog - 128 ;
						if( (uint )(analog+8) <= 16 ) {
							analog = 0 ;
						}

						analog += (analog / 2) ;

						if( analog < -128 ) {
							analog = -128 ;
						} else
						if( analog >  127 ) {
							analog =  127 ;
						}

						g.pad.paddle.pos = (uint8_t )(analog + 128) ;
					}
					data = 0xc0 | (g.pad.paddle.pos & 15) ;
				} else {
					data = 0xe0 | (g.pad.paddle.pos >> 4) ;
				}

				data |= (((((g.pad.norm>>(PAD_BTN_+0))|(g.pad.norm>>(PAD_BTN_+1)))&1)^1) << 4) ;

				g.pad.paddle.pdsel = (g.pad.paddle.pdsel + 1) & 15 ;

			}
			break ;
		}
	}

	return data ;
}

static uint8_t RAMFUNC addressbus_read_mem(ADDRESSBUS *p_this, uint16_t address)
{
	return m_sm3gg.memInfo[address/0x2000].p_read[address&0x1fff] ;
}

static uint8_t RAMFUNC addressbus_read_mem_bank(ADDRESSBUS *p, uint16_t address)
{
	if( address < 0x0400 ) {
		return m_sm3gg.p_rom[address] ;
	}

	return m_sm3gg.memInfo[address/0x2000].p_read[address&0x1fff] ;
}


// -------------------------------------------------------------------------------- 
//!  アドレスバス書き込み                                                           
//!                                                                                 
//!  @param [in,out] p_this   this ポインタ                                         
//!  @param          address  アドレス                                              
//!  @param          data     データ                                                
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
static void RAMFUNC addressbus_write_io(ADDRESSBUS *p_this, uint16_t address, uint8_t data)
{
	switch( (address/0x40) & 3 ) {
		case 0:		// 0x00 - 0x3f
		{
			if( g.mode == MODE_SM3 ) {
				return ;
			}

			switch( address ) {
				case 0x01:
				{
					m_sm3gg.gg.extData = data ;
					break ;
				}

				case 0x02:
				{
					m_sm3gg.gg.extSet = data ;
					break ;
				}

				case 0x03:
				{
					m_sm3gg.gg.sendData = data ;
					break ;
				}

				case 0x05:
				{
					m_sm3gg.gg.comMode = data ;
					break ;
				}

//				case 0x00:
//				case 0x04:
//				case 0x06:
//				case 0x30:
//				case 0x31:
				default:
					break ;
			}
			break ;
		}

		case 1:		// 0x40 - 0x7f
		{
			sound_writeReg( data ) ;
			break ;
		}

		case 2:		// 0x80 - 0xbf
		{
			writeAddressBus(&m_sm3gg.vdp.addressBus, address&1, data) ;
			break ;
		}

		default:	// 0xc0 - 0xff
		{
			break ;
		}
	}
}

static void RAMFUNC addressbus_write_mem(ADDRESSBUS *p, uint16_t address, uint8_t data)
{
	uint8_t		*p_ram ;


	p_ram = &m_sm3gg.memInfo[address/0x2000].p_write[0] ;
	if( p_ram != NULL ) {
		p_ram[address&0x1fff] = data ;
	}
}

static void RAMFUNC addressbus_write_mem_bank(ADDRESSBUS *p, uint16_t address, uint8_t data)
{
	uint8_t		*p_ram ;


	if( address >= 0xfffc ) {
		m_sm3gg.bank[address&3] = data ;
		switch( address ) {
			case 0xfffc:
			case 0xffff:
			{
				if( m_sm3gg.bank[0] & (1<<3) ) {	// RAM
					setRamBank(0x8000, ((m_sm3gg.bank[0]>>2)&1)*0x4000) ;
				} else {							// ROM
					setRomBank(0x8000,   m_sm3gg.bank[3]       *0x4000) ;
				}
				break ;
			}

			case 0xfffd:
			{
				setRomBank(0x0000, data*0x4000) ;
				break ;
			}

			case 0xfffe:
			{
				setRomBank(0x4000, data*0x4000) ;
				break ;
			}
		}
	}

	p_ram = &m_sm3gg.memInfo[address/0x2000].p_write[0] ;
	if( p_ram != NULL ) {
		p_ram[address&0x1fff] = data ;
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void sm3gg_init( void )
{
	memset(&m_sm3gg, 0, sizeof( m_sm3gg )) ;
	m_sm3gg.addressBusIo.read   = addressbus_read_io ;
	m_sm3gg.addressBusIo.write  = addressbus_write_io ;
	m_sm3gg.addressBusMem.read  = addressbus_read_mem ;
	m_sm3gg.addressBusMem.write = addressbus_write_mem ;

	z80_init(&m_sm3gg.z80, &m_sm3gg.addressBusIo, &m_sm3gg.addressBusMem) ;
	vdp_init(&m_sm3gg.vdp, &m_sm3gg.z80.pinInt) ;

	z80_reset( &m_sm3gg.z80 ) ;
	vdp_reset( &m_sm3gg.vdp ) ;
}


// -------------------------------------------------------------------------------- 
//!  ROM 情報を設定                                                                 
//!                                                                                 
//!  @param [in] p_romInfo  ROM 情報                                                
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void sm3gg_setRomInfo( ROMINFO const *p_romInfo )
{
	g.mode = p_romInfo->mode ;

	m_sm3gg.romMask = p_romInfo->romSize - 1 ;
	if( p_romInfo->romSize == (48*1024) ) {
		m_sm3gg.romMask = 65536 - 1 ;
	}

	switch( p_romInfo->ramSize / 0x20 ) {
		case 0:		// 0kB
		case 1:		// 8kB
		{
			m_sm3gg.ramBase[0] = 0x0000 ;
			m_sm3gg.ramBase[1] = 0x0000 ;
			m_sm3gg.ramBase[2] = 0x0000 ;
			m_sm3gg.ramBase[3] = 0x0000 ;
			break ;
		}

		case 2:		// 16kB
		{
			m_sm3gg.ramBase[0] = 0x0000 ;
			m_sm3gg.ramBase[1] = 0x2000 ;
			m_sm3gg.ramBase[2] = 0x0000 ;
			m_sm3gg.ramBase[3] = 0x2000 ;
			break ;
		}

		case 3:		// 24kB
		{
			m_sm3gg.ramBase[0] = 0x0000 ;
			m_sm3gg.ramBase[1] = 0x2000 ;
			m_sm3gg.ramBase[2] = 0x4000 ;
			m_sm3gg.ramBase[3] = 0x4000 ;
			break ;
		}

		default:	// 32kB
		{
			m_sm3gg.ramBase[0] = 0x0000 ;
			m_sm3gg.ramBase[1] = 0x2000 ;
			m_sm3gg.ramBase[2] = 0x4000 ;
			m_sm3gg.ramBase[3] = 0x6000 ;
			break ;
		}
	}

	m_sm3gg.p_rom              = p_romInfo->p_rom ;
	m_sm3gg.memInfo[6].p_read  = &m_sm3gg.ram[0] ;
	m_sm3gg.memInfo[6].p_write = &m_sm3gg.ram[0] ;
	m_sm3gg.memInfo[7].p_read  = &m_sm3gg.ram[0] ;
	m_sm3gg.memInfo[7].p_write = &m_sm3gg.ram[0] ;

	if( p_romInfo->romSize <= (48*1024) ) {		// バンク管理なし
		m_sm3gg.addressBusMem.read  = addressbus_read_mem ;
		m_sm3gg.addressBusMem.write = addressbus_write_mem ;
	} else {									// バンク管理あり
		m_sm3gg.addressBusMem.read  = addressbus_read_mem_bank ;
		m_sm3gg.addressBusMem.write = addressbus_write_mem_bank ;
	}

	sm3gg_reset() ;
}


// -------------------------------------------------------------------------------- 
//!  リセット                                                                       
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void sm3gg_reset( void )
{
	g.p_clockEventList  = NULL ;
	m_sm3gg.bank[0]     = 0x00 ;
	m_sm3gg.bank[1]     = 0x00 ;
	m_sm3gg.bank[2]     = 0x01 ;
	m_sm3gg.bank[3]     = 0x02 ;
	m_sm3gg.gg.extData  = 0xff ;
	m_sm3gg.gg.extSet   = 0xff ;
	m_sm3gg.gg.sendData = 0x00 ;
	m_sm3gg.gg.comMode  = 0x00 ;

	setRomBank(0x0000, 0x0000) ;
	setRomBank(0x4000, 0x4000) ;
	setRomBank(0x8000, 0x8000) ;

	writePin(&m_sm3gg.z80.pinInt, 1) ;

	z80_reset( &m_sm3gg.z80 ) ;
	vdp_reset( &m_sm3gg.vdp ) ;
}


// -------------------------------------------------------------------------------- 
//!  クロック供給                                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sm3gg_supplyClock( void )
{
	// PAUSE
	if( g.mode == MODE_SM3 ) {
		writePin(&m_sm3gg.z80.pinNmi, (~g.pad.norm>>PAD_START_)&1) ;
	}

	clockgen_supplyClock() ;
}
