// -------------------------------------------------------------------------------- 
//!  @file  sound.c                                                                 
//!  サウンドのソースファイル                                                       
//                                                                                  
//   Copyright (C) 2022 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"

#ifdef  SOUNDDAC_CFG_DAC_MCP49x1
#define DAC_INIT					mcp49x1_init					//!< DAC 初期化処理
#define DAC_WRITE					mcp49x1_write					//!< DAC 書き込み処理
#define DAC_WAIT					mcp49x1_wait					//!< DAC 書き込み待ち処理
#endif//SOUNDDAC_CFG_DAC_MCP49x1


	enum {
		SAMPLEFREQ					=	1000000 / SOUNDDAC_CFG_TIME,//!< サンプリング周波数
		UPDATESAMPLE				=	SAMPLEFREQ / 100,			//!< 更新サンプル数
		BASECLOCK					=	10738635,					//!< ベースクロック
		UPDATECLOCK					=	(BASECLOCK + (SAMPLEFREQ-1)) / SAMPLEFREQ,
																	//!< 更新クロック
	} ;

	//! レジスタ情報
	typedef struct {
		uint16_t					pos ;							//!< 位置
		uint16_t					reg[4] ;						//!< レジスタ
	} REGINFO ;

	//! サウンド
	typedef struct {
		uint8_t						mute ;							//!< ミュート状態
		uint8_t						reg ;							//!< レジスタ番号
		uint8_t						readPos  ;						//!< レジスタ情報読み取り位置
		uint8_t						writePos ;						//!< レジスタ情報書き込み位置
		uint16_t					output ;						//!< 出力値
		uint16_t					playPos ;						//!< 再生位置
		uint16_t					startPos ;						//!< 開始再生位置
		uint32_t					startClock ;					//!< 開始クロック
		REGINFO						*p_regInfo ;					//!< カレントレジスタ情報
		PCMSN76489INFO				info ;							//!< SN76489 情報
		struct repeating_timer		timer ;							//!< 繰り返しタイマー
		REGINFO						regInfo[256] ;					//!< レジスタ情報
	} SOUND ;


static SOUND						m_sound ;						//!< サウンド


// -------------------------------------------------------------------------------- 
//!  繰り返しタイマー割り込みサービスルーチン                                       
//!                                                                                 
//!  @param [in] p_timer  繰り返しタイマー                                          
//!  @retval     true     繰り返し                                                  
//!  @retval     false    停止                                                      
// -------------------------------------------------------------------------------- 
static bool RAMFUNC isr_timer( struct repeating_timer *p_timer )
{
	REGINFO		*p_regInfo ;
	uint16_t	reg3 ;


	DAC_WRITE( m_sound.output ) ;

	m_sound.output = (uint16_t )(pcm_sn76489_update( &m_sound.info ) + 0x0800) ;

	if( m_sound.readPos != m_sound.writePos ) {
		p_regInfo = &m_sound.regInfo[m_sound.readPos] ;

		if( m_sound.playPos == p_regInfo->pos ) {
			m_sound.readPos++ ;

			reg3 = (p_regInfo->reg[3] & 0xf000) | ((p_regInfo->reg[3] & 0x0004) << 8) | 0x0800 ;
			if( (p_regInfo->reg[3]&3) != 3 ) {	// SHIFT
				reg3 |= (32 << (p_regInfo->reg[3] & 3)) ;
			} else {							// Tone Generator
				reg3 |= (p_regInfo->reg[2] & 0x03ff) ;
			}
			pcm_sn76489_setReg(&m_sound.info, 0, p_regInfo->reg[0]) ;
			pcm_sn76489_setReg(&m_sound.info, 1, p_regInfo->reg[1]) ;
			pcm_sn76489_setReg(&m_sound.info, 2, p_regInfo->reg[2]) ;
			pcm_sn76489_setReg(&m_sound.info, 3, reg3) ;
		}
	}

	DAC_WAIT() ;

	m_sound.playPos++ ;

	if( m_sound.mute ) {
		m_sound.output = 0x0800 ;
	}

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  レジスタ情報を更新                                                             
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC updateRegInfo( void )
{
	REGINFO		*p_regInfo ;
	uint16_t	pos ;


	pos       = m_sound.startPos + ((g.baseClock - m_sound.startClock) / UPDATECLOCK) ;
	p_regInfo = m_sound.p_regInfo ;

	if( m_sound.readPos != m_sound.writePos ) {
		if( pos == p_regInfo->pos ) {
			return ;
		}
	}

	m_sound.p_regInfo         = &m_sound.regInfo[m_sound.writePos] ;
	m_sound.p_regInfo->pos    = pos ;
	m_sound.p_regInfo->reg[0] = p_regInfo->reg[0] ;
	m_sound.p_regInfo->reg[1] = p_regInfo->reg[1] ;
	m_sound.p_regInfo->reg[2] = p_regInfo->reg[2] ;
	m_sound.p_regInfo->reg[3] = p_regInfo->reg[3] ;

	m_sound.writePos++ ;
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void sound_init( void )
{
	int		index ;


	DAC_INIT() ;

	sound_reset() ;

	for(index=0; index<=0x0800; index+=0x0040) {
		DAC_WRITE( index ) ;
		DAC_WAIT() ;
		sleep_us( 10 ) ;
	}

	add_repeating_timer_us(-SOUNDDAC_CFG_TIME, isr_timer, NULL, &m_sound.timer) ;
}


// -------------------------------------------------------------------------------- 
//!  リセット                                                                       
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void sound_reset( void )
{
	__disable_irq() ;

	m_sound.reg               = 0 ;
	m_sound.output            = 0x0800 ;
	m_sound.readPos           = 0 ;
	m_sound.writePos          = 0 ;
	m_sound.p_regInfo         = &m_sound.regInfo[0] ;
	m_sound.regInfo[0].reg[0] = 0xf000 ;
	m_sound.regInfo[0].reg[1] = 0xf000 ;
	m_sound.regInfo[0].reg[2] = 0xf000 ;
	m_sound.regInfo[0].reg[3] = 0xf000 ;

	pcm_sn76489_reset( &m_sound.info ) ;

	__enable_irq() ;
}


// -------------------------------------------------------------------------------- 
//!  レジスタ書き込み                                                               
//!                                                                                 
//!  @param [in] data  データ                                                       
//!  @return           なし                                                         
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_writeReg( uint8_t data )
{
	uint16_t	value, *p_reg ;


	updateRegInfo() ;

	if( data & 0x80 ) {
		m_sound.reg = (uint8_t )((data >> 4) & 7) ;
	}

	p_reg = &m_sound.p_regInfo->reg[m_sound.reg/2] ;

	if( !(m_sound.reg&1) ) {	// 周波数 / ノイズコントロール

		if( data & 0x80 ) {
			value = (*p_reg & 0xfff0) |  (uint16_t )(data & 0x0f) ;
		} else {
			value = (*p_reg & 0xf00f) | ((uint16_t )(data & 0x3f) << 4) ;
		}

	} else {					// アッテネータ

		if( (data&0x80) || (m_sound.reg==7) ) {
			value = (*p_reg & 0x0fff) | ((uint16_t )(data & 0x0f) << 12) ;
		} else {
			value = (*p_reg & 0xf000) |  (uint16_t )(data & 0x0f) ;
		}

	}

	*p_reg = value ;
}


// -------------------------------------------------------------------------------- 
//!  消音                                                                           
//!                                                                                 
//!  @param [in] mute  true:ミュート, false:ミュート解除                            
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_mute( bool mute )
{
	m_sound.mute = (uint8_t )mute ;
}


// -------------------------------------------------------------------------------- 
//!  更新                                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_update( void )
{
	m_sound.startClock = g.baseClock ;
	m_sound.startPos   = m_sound.playPos + UPDATESAMPLE ;
}
