// -------------------------------------------------------------------------------- 
//!  @file  wavecut.c                                                               
//!  wave ファイルの前後無音削除ツール                                              
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	enum {
		CHUNK_ID_RIFF				=	0x46464952,					//!< RIFF チャンク ID
		CHUNK_ID_FMT				=	0x20746d66,					//!< fmt  チャンク ID
		CHUNK_ID_DATA				=	0x61746164,					//!< data チャンク ID
		RIFF_FORMAT					=	0x45564157,					//!< RIFF チャンクフォーマット
	} ;

	//! ファイル情報
	typedef struct {
		FILE						*p_file ;						//!< ファイルポインタ
		uint32_t					size ;							//!< サイズ
		uint32_t					pos ;							//!< 位置
	} FILEINFO ;

	//! チャンク
	typedef struct {
		uint32_t					id ;							//!< チャンク ID
		uint32_t					size ;							//!< チャンクサイズ
	} CHUNK ;

	//! 波形フォーマット
	typedef struct {
		uint16_t					formatTag ;						//!< フォーマットタグ
		uint16_t					channels ;						//!< チャンネル数
		uint32_t					samplesPerSec ;					//!< サンプリング周波数
		uint32_t					avgBytesPerSec ;				//!< １秒あたりのバイト数
		uint16_t					blockAlign ;					//!< ブロックサイズ
		uint16_t					bitsPerSample ;					//!< サンプルビット数
	} WAVEFORMAT ;


// -------------------------------------------------------------------------------- 
//!  ファイル読み込み                                                               
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @param [out]    p_buff      読み込みバッファアドレス                           
//!  @param [in]     size        読み込みバッファサイズ                             
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int readFile(FILEINFO *p_fileInfo, void *p_buff, uint32_t size)
{
	if( (p_fileInfo->pos+size) > p_fileInfo->size ) {
		return -1 ;
	}

	if( fread(p_buff, size, 1, p_fileInfo->p_file) != 1 ) {
		return -1 ;
	}

	p_fileInfo->pos += size ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ファイルシーク                                                                 
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @param [in]     pos         シーク位置                                         
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int seekFile(FILEINFO *p_fileInfo, uint32_t pos)
{
	if( p_fileInfo->pos == pos ) {
		return 0 ;
	}

	if( p_fileInfo->size < pos ) {
		return -1 ;
	}

	if( fseek(p_fileInfo->p_file, pos, SEEK_SET) == -1 ) {
		return -1 ;
	}

	p_fileInfo->pos = pos ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  RIFF チャンクを確認                                                            
//!                                                                                 
//!  @param [in,out] p_fileInfo  ファイル情報                                       
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int checkRiffChunk( FILEINFO *p_fileInfo )
{
	CHUNK		chunk ;
	uint32_t	format ;


	if( p_fileInfo->size < sizeof( CHUNK ) ) {
		printf("サポートしていないファイル形式です。\n") ;
		return -1 ;
	}

	// RIFF チャンク読み込み
	if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
		printf("RIFFチャンクの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( chunk.id != CHUNK_ID_RIFF ) {
		printf("RIFFチャンクがありません。\n") ;
		return -1 ;
	}

	if( chunk.size < (p_fileInfo->size-sizeof( CHUNK )) ) {
		printf("RIFFチャンクのサイズが正しくありません。\n") ;
		return -1 ;
	}

	p_fileInfo->size = sizeof( CHUNK ) + chunk.size ;

	if( readFile(p_fileInfo, &format, sizeof( format )) ) {
		printf("RIFFチャンクのフォーマットの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( format != 0x45564157 ) {		// 0x45564157 = "WAVE"
		printf("サポートしていないファイル形式です。\n") ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  波形フォーマットを読み込む                                                     
//!                                                                                 
//!  @param [in,out] p_fileInfo    ファイル情報                                     
//!  @param [out]    p_waveFormat  波形フォーマット格納アドレス                     
//!  @retval          0            正常終了                                         
//!  @retval         -1            異常終了                                         
// -------------------------------------------------------------------------------- 
static int readWaveFormat(FILEINFO *p_fileInfo, WAVEFORMAT *p_waveFormat)
{
	CHUNK	chunk ;
	int		remain ;

	// fmt チャンク読み込み
	if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
		printf("fmt チャンクの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	if( chunk.id != 0x20746d66 ) {		// 0x20746d66 = "fmt "
		printf("fmt チャンクがありません。\n") ;
		return -1 ;
	}

	if( chunk.size < sizeof( WAVEFORMAT ) ) {
		printf("fmt チャンクのサイズが正しくありません。\n") ;
		return -1 ;
	}

	if( readFile(p_fileInfo, p_waveFormat, sizeof( WAVEFORMAT )) ) {
		printf("フォーマットの読み込みに失敗しました。\n") ;
		return -1 ;
	}

	remain = ((chunk.size + 1) & ~1) - sizeof( WAVEFORMAT ) ;
	if( remain ) {
		if( seekFile(p_fileInfo, p_fileInfo->pos+remain) ) {
			printf("フォーマットの読み込みに失敗しました。\n") ;
			return -1 ;
		}
	}

	printf("fmt:\n") ;
	printf("    formatTag      = 0x%04x\n", p_waveFormat->formatTag) ;
	printf("    channels       = %d\n",     p_waveFormat->channels) ;
	printf("    samplesPerSec  = %d\n",     p_waveFormat->samplesPerSec) ;
	printf("    avgBytesPerSec = %d\n",     p_waveFormat->avgBytesPerSec) ;
	printf("    blockAlign     = %d\n",     p_waveFormat->blockAlign) ;
	printf("    bitsPerSample  = %d\n",     p_waveFormat->bitsPerSample) ;

	if( p_waveFormat->formatTag != 0x0001 ) {
		printf("PCM 形式以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( p_waveFormat->channels != 1 ) {
		printf("モノラル以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( p_waveFormat->bitsPerSample != 8 ) {
		printf("8ビット以外はサポートしていません。\n") ;
		return -1 ;
	}

	if( !p_waveFormat->samplesPerSec || (p_waveFormat->avgBytesPerSec!=p_waveFormat->samplesPerSec) || (p_waveFormat->blockAlign!=1) ) {
		printf("フォーマットが正しくありません。\n") ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  波形データを読み込む                                                           
//!                                                                                 
//!  @param [in,out] p_fileInfo    ファイル情報                                     
//!  @param [in]     p_waveFormat  波形フォーマット                                 
//!  @param [out]    pp_waveData   波形データアドレスの格納アドレス                 
//!  @retval          0 以上       波形データサイズ                                 
//!  @retval         -1            異常終了                                         
// -------------------------------------------------------------------------------- 
static int readWaveData(FILEINFO *p_fileInfo, WAVEFORMAT *p_waveFormat, uint8_t **pp_waveData)
{
	CHUNK		chunk ;
	uint8_t		*p_waveData ;


	while( 1 ) {
		// チャンク読み込み
		if( readFile(p_fileInfo, &chunk, sizeof( chunk )) ) {
			printf("dataチャンクの読み込みに失敗しました。\n") ;
			return -1 ;
		}

		if( chunk.id == CHUNK_ID_DATA ) {
			break ;
		}

		if( seekFile(p_fileInfo, p_fileInfo->pos+((chunk.size+1)&~1)) ) {
			printf("dataチャンクの読み込みに失敗しました。\n") ;
			return -1 ;
		}
	}

	if( !chunk.size || ((p_fileInfo->pos+chunk.size)>p_fileInfo->size) || (chunk.size%p_waveFormat->blockAlign) ) {
		printf("dataチャンクのサイズ（%d）が正しくありません。\n", chunk.size) ;
		return -1 ;
	}

	// 波形データ確保
	p_waveData = malloc( chunk.size ) ;
	if( p_waveData == NULL ) {
		printf("メモリが不足しています\n") ;
		return -1 ;
	}

	if( readFile(p_fileInfo, &p_waveData[0], chunk.size) ) {
		printf("dataチャンクの読み込みに失敗しました。\n") ;
		free( p_waveData ) ;
		return -1 ;
	}

	printf("\ndata:\n") ;
	printf("    sample = %d\n", chunk.size/p_waveFormat->blockAlign) ;

	*pp_waveData = &p_waveData[0] ;

	return (chunk.size/p_waveFormat->blockAlign) ;
}


// -------------------------------------------------------------------------------- 
//!  波形データをロード                                                             
//!                                                                                 
//!  @param [in]  p_fileName    ファイル名                                          
//!  @param [out] p_waveFormat  波形フォーマット格納アドレス                        
//!  @param [out] pp_wave       波形データアドレスの格納アドレス                    
//!  @retval       0 以上       サンプル数                                          
//!  @retval      -1            異常終了                                            
// -------------------------------------------------------------------------------- 
static int loadWaveData(char const *p_fileName, WAVEFORMAT *p_waveFormat, uint8_t **pp_waveData)
{
	FILEINFO	fileInfo ;
	int			result ;
	long		fileSize ;

	// ファイル開く
	fileInfo.p_file = fopen(&p_fileName[0], "rb") ;
	if( fileInfo.p_file == NULL ) {
		fprintf(stderr, "%sファイルを開くのに失敗しました。\n", &p_fileName[0]) ;
		return -1 ;
	}

	result = -1 ;

	do {
		// ファイルサイズ取得
		fseek(fileInfo.p_file, 0, SEEK_END) ;

		fileSize = ftell( fileInfo.p_file ) ;
		if( fileSize <= 0 ) {
			break ;
		}
		fileInfo.size = (uint32_t )fileSize ;

		fseek(fileInfo.p_file, 0, SEEK_SET) ;

		fileInfo.pos = 0 ;

		// RIFF チャンク確認
		if( checkRiffChunk( &fileInfo ) ) {
			break ;
		}

		// 波形フォーマット読み込み
		if( readWaveFormat(&fileInfo, p_waveFormat) ) {
			break ;
		}

		// 波形データ読み込み
		result = readWaveData(&fileInfo, p_waveFormat, pp_waveData) ;
		if( result < 0 ) {
			break ;
		}
	} while( 0 ) ;

	fclose( fileInfo.p_file ) ;

	return result ;
}


// -------------------------------------------------------------------------------- 
//!  波形データの先頭有音位置を取得                                                 
//!                                                                                 
//!  @param [in] p_waveData    波形データ                                           
//!  @param [in] sample        サンプル数                                           
//!  @param [in] p_waveFormat  波形フォーマット                                     
//!  @return                   サンプル位置                                         
// -------------------------------------------------------------------------------- 
static int getWaveDataHead(uint8_t *p_waveData, int sample, WAVEFORMAT *p_waveFormat)
{
	int16_t		*p_waveData16 ;
	int			index, count ;


	if( p_waveFormat->bitsPerSample == 8 ) {

		for(index=0; index<sample; index++) {
			for(count=0; count<p_waveFormat->channels; count++) {
				if( *p_waveData++ != 0x80 ) {
					return index ;
				}
			}
		}

	} else {

		p_waveData16 = (int16_t *)&p_waveData[0] ;

		for(index=0; index<sample; index++) {
			for(count=0; count<p_waveFormat->channels; count++) {
				if( *p_waveData16++ ) {
					return index ;
				}
			}
		}

	}

	return sample ;
}


// -------------------------------------------------------------------------------- 
//!  波形データの終了有音位置を取得                                                 
//!                                                                                 
//!  @param [in] p_waveData    波形データ                                           
//!  @param [in] sample        サンプル数                                           
//!  @param [in] p_waveFormat  波形フォーマット                                     
//!  @return                   サンプル位置                                         
// -------------------------------------------------------------------------------- 
static int getWaveDataTail(uint8_t *p_waveData, int sample, WAVEFORMAT *p_waveFormat)
{
	int16_t		*p_waveData16 ;
	int			index, count ;


	p_waveData += ((sample - 1) * p_waveFormat->blockAlign) ;

	if( p_waveFormat->bitsPerSample == 8 ) {

		for(index=sample-1; index>=0; index--) {
			for(count=0; count<p_waveFormat->channels; count++) {
				if( *p_waveData-- != 0x80 ) {
					return index ;
				}
			}
		}

	} else {

		p_waveData16 = (int16_t *)&p_waveData[0] ;

		for(index=sample-1; index>=0; index--) {
			for(count=0; count<p_waveFormat->channels; count++) {
				if( *p_waveData16-- ) {
					return index ;
				}
			}
		}

	}

	return sample ;
}


// -------------------------------------------------------------------------------- 
//!  波形ファイルを作成                                                             
//!                                                                                 
//!  @param [in] p_fileName    ファイル名                                           
//!  @param [in] p_waveData    波形データ                                           
//!  @param [in] sample        サンプル数                                           
//!  @param [in] p_waveFormat  波形フォーマット                                     
//!  @retval      0            正常終了                                             
//!  @retval     -0            異常終了                                             
// -------------------------------------------------------------------------------- 
static int createWaveFile(char const *p_fileName, uint8_t *p_waveData, int sample, WAVEFORMAT *p_waveFormat)
{
	FILE		*p_file ;
	CHUNK		chunk ;
	WAVEFORMAT	waveFormat ;
	uint32_t	data ;
	int			result, size ;


	p_file = fopen(&p_fileName[0], "wb") ;
	if( p_file == NULL ) {
		return -1 ;
	}

	result = -1 ;
	size   = sample * p_waveFormat->blockAlign ;

	do {
		// RIFF チャンク書き込み
		chunk.id   = CHUNK_ID_RIFF ;
		chunk.size = sizeof( data ) +
		             sizeof( chunk ) + sizeof( waveFormat ) +
		             sizeof( chunk ) + size ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}

		data = RIFF_FORMAT ;
		if( fwrite(&data, sizeof( data ), 1, p_file) != 1 ) {
			break ;
		}

		// fmt チャンク書き込み
		chunk.id   = CHUNK_ID_FMT ;
		chunk.size = sizeof( WAVEFORMAT ) ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}
		if( fwrite(p_waveFormat, sizeof( WAVEFORMAT ), 1, p_file) != 1 ) {
			break ;
		}

		// data チャンク書き込み
		chunk.id   = CHUNK_ID_DATA ;
		chunk.size = size ;
		if( fwrite(&chunk, sizeof( chunk ), 1, p_file) != 1 ) {
			break ;
		}

		if( fwrite(&p_waveData[0], size, 1, p_file) != 1 ) {
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	result |= fclose( p_file ) ;

	if( result ) {
		remove( &p_fileName[0] ) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  メイン処理                                                                     
//!                                                                                 
//!  @param [in] argc  引数の数                                                     
//!  @param [in] argv  引数                                                         
//!  @retval     0     正常終了                                                     
//!  @retval     1     異常終了                                                     
//!  @retval     2     ヘルプ表示                                                   
// -------------------------------------------------------------------------------- 
int main(int argc, char *argv[])
{
	WAVEFORMAT	waveFormat ;
	uint8_t		*p_waveData ;
	int			result, sample, head, tail ;


	if( argc < 3 ) {
		fprintf(stderr, "usage: wavecus srcefile destfile\n") ;
		return 2 ;
	}

	sample = loadWaveData(argv[1], &waveFormat, &p_waveData) ;
	if( sample <= 0 ) {
		return 1 ;
	}

	result = 1 ;

	do {
		if( (waveFormat.bitsPerSample!=8) && (waveFormat.bitsPerSample!=16) ) {
			fprintf(stderr, "8/16ビット以外の波形フォーマットに対応していません。\n") ;
			break ;
		}

		head = getWaveDataHead(&p_waveData[0], sample, &waveFormat) ;
		if( head == sample ) {
			fprintf(stderr, "波形データがありません。\n") ;
			break ;
		}

		tail = getWaveDataTail(&p_waveData[0], sample, &waveFormat) ;

		if( head ) {
			head-- ;
		}

		if( tail != (sample-1) ) {
			tail++ ;
		}

		if( !head && (tail==(sample-1)) ) {
			fprintf(stderr, "削除できる無音はありません。\n") ;
			break ;
		}

		if( createWaveFile(argv[2], &p_waveData[head*waveFormat.blockAlign], tail-head+1, &waveFormat) ) {
			fprintf(stderr, "%s ファイルの生成に失敗しました。\n", argv[2]) ;
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	free( p_waveData ) ;

	return result ;
}
