// -------------------------------------------------------------------------------- 
//!  @file  common.h                                                                
//!  共通処理・データのヘッダファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#ifndef __COMMON_H__INCLUDED__
#define __COMMON_H__INCLUDED__


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pico/stdlib.h>
#include "config.h"
#include "libvga/vga/vga.h"
#include "libvga/pad/pad.h"
#include "data.h"
#include "sound.h"


#ifdef  DEBUG
#define DBGASSERT( e )				if( !(e) ) { printf("******** %s(%d)\n", __func__, __LINE__) ;  while( 1 ) {} }
																	//!< デバッグ用アサート
#define DBGTRACE(fmt, ...)			printf(fmt, __VA_ARGS__)		//!< デバッグ用トレース
#else
#define DBGASSERT( e )												//!< デバッグ用アサート
#define DBGTRACE(fmt, ...)											//!< デバッグ用トレース
#endif//DEBUG

#define MEMBERSTRUCT(_s, _m, _p)	((_s *)((uintptr_t )(_p) - (uintptr_t )&((_s *)0)->_m))
																	//!< メンバーアドレスから構造体の先頭アドレスを取得

#define __enable_irq()				__asm volatile ("cpsie i")		//!< 割り込み許可
#define __disable_irq()				__asm volatile ("cpsid i")		//!< 割り込み禁止


#define WALL						VGATXT_OP						//!< 壁当たり


	enum {
		DMA_CH						=	2,							//!< DMA チャンネル
		SCREEN_WIDTH				=	VGA_TXT_WIDTH_COUNT,		//!< 画面幅
		SCREEN_HEIGHT				=	VGA_TXT_HEIGHT_COUNT,		//!< 画面高さ
		DRAW_WIDTH					=	SCREEN_WIDTH,				//!< 描画幅
		DRAW_HEIGHT					=	SCREEN_HEIGHT-2,			//!< 描画高さ
		TASK_MAX					=	128,						//!< タスク最大数
		FP_ONE_						=	8,							//!< 固定小数点１ビット数
		FP_ONE						=	1 << FP_ONE_,				//!< 固定小数点１
	} ;

	//! 8bit 固定小数点
	typedef union {
		int16_t						n ;								//!< 数値
		struct {
			uint8_t					l ;								//!< 小数部
			int8_t					h ;								//!< 整数部
		} ;
	} FI8 ;

#include "task.h"
#include "obj.h"

	//! フォースビーム
	typedef struct {
		uint8_t						remain ;						//!< ビーム残り数
		int8_t						rbx ;							//!< 反射レーザー開始ｘ座標
		int8_t						rby ;							//!< 反射レーザー開始ｙ座標
	} FORCEBEAM ;

	//! グローバル変数
	typedef struct {
		int16_t						sinData[FP_ONE] ;				//!< サインデータ
		uint8_t						angleData[DRAW_HEIGHT][DRAW_WIDTH] ;
																	//!< 角度データ
		uint16_t					count1 ;						//!< １フレームカウンタ

		//! VRAM
		struct {
			uint32_t				pat[256][32/4] ;				//!< パターンデータ
			uint16_t				txt[SCREEN_HEIGHT][SCREEN_WIDTH] ;
																	//!< テキストデータ
		} vram ;

		//! パッド状態
		struct {
			uint16_t				norm ;							//!< 通常
			uint16_t				old ;							//!< 以前
			uint16_t				once ;							//!< １度のみ
		} pad ;

		//! 背景
		struct {
			FI8						x ;								//!< ｘ座標
			FI8						y ;								//!< ｙ座標
			int8_t					xofs ;							//!< ｘ座標移動オフセット
			int8_t					yofs ;							//!< ｙ座標移動オフセット
		} bg ;

		TCBLIST						playerList ;					//!< プレイヤーリスト
		TCBLIST						enemyList ;						//!< 敵リスト
		TCB							playerTcb ;						//!< プレイヤータスクのTCB
		TCB							forceTcb ;						//!< フォースタスクの TCB
		FORCEBEAM					forceBeam[4] ;					//!< フォースビーム
		uint8_t						bitCount ;						//!< ビット数
	} GLOBALVAR ;


extern GLOBALVAR					g ;								//!< グローバル変数


extern void            common_init( void ) ;
extern void    RAMFUNC setMemory(void *s, uint8_t c, int n) ;
extern void    RAMFUNC copyMemory(void *d, void const *s, int n) ;
extern uint8_t RAMFUNC getAngle(int sx, int sy, int ex, int ey) ;


// -------------------------------------------------------------------------------- 
//!  数値から固定小数点へ変換                                                       
//!                                                                                 
//!  @param [in] n  数値                                                            
//!  @return        固定小数点                                                      
// -------------------------------------------------------------------------------- 
#define toFp( n )					(FP_ONE*(n))


// -------------------------------------------------------------------------------- 
//!  サインを取得                                                                   
//!                                                                                 
//!  @param [in] angle  角度                                                        
//!  @return            固定小数のサイン値                                          
// -------------------------------------------------------------------------------- 
static inline FI8 getSin( uint8_t angle )
{
	return (FI8 )g.sinData[angle] ;
}


// -------------------------------------------------------------------------------- 
//!  コサインを取得                                                                 
//!                                                                                 
//!  @param [in] angle  角度                                                        
//!  @return            固定小数のコサイン値                                        
// -------------------------------------------------------------------------------- 
static inline FI8 getCos( uint8_t angle )
{
	return (FI8 )g.sinData[(uint8_t )(angle+64)] ;
}


#endif//__COMMON_H__INCLUDED__
