// -------------------------------------------------------------------------------- 
//!  @file  obj.c                                                                   
//!  オブジェクト管理のソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"


// -------------------------------------------------------------------------------- 
//!  アクションを設定                                                               
//!                                                                                 
//!  @param [in,out] p_tcb   TCB                                                    
//!  @param [in]     action  アクション番号                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_setAction(TCB *p_tcb, uint8_t action)
{
	p_tcb->status      &= ~TCB_STATUS_ACTEND ;
	p_tcb->p_action     = g_p_actData[action] ;
	p_tcb->pattern      = p_tcb->p_action[0] ;
	p_tcb->patternCount = p_tcb->p_action[1] ;
}


// -------------------------------------------------------------------------------- 
//!  アクションを更新                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  TCB                                                     
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
void RAMFUNC obj_updateAction( TCB *p_tcb )
{
	const uint8_t	*p_action ;
	uint8_t			pattern, count ;

	// アクション終了
	if( p_tcb->status & TCB_STATUS_ACTEND ) {
		return ;
	}

	if( --p_tcb->patternCount ) {
		return ;
	}

	p_action = &p_tcb->p_action[2] ;
	pattern  = p_action[0] ;
	count    = p_action[1] ;

	if( !count ) {

		// 終了
		if( !pattern ) {
			p_tcb->status |= TCB_STATUS_ACTEND ;
			return ;
		}

		// ループ
		p_action -= (pattern * 2) ;
		pattern   = p_action[0] ;
		count     = p_action[1] ;
	}

	p_tcb->p_action     = &p_action[0] ;
	p_tcb->pattern      = pattern ;
	p_tcb->patternCount = count ;
}


// -------------------------------------------------------------------------------- 
//!  オブジェクトを描画                                                             
//!                                                                                 
//!  @param [in] x        ｘ座標                                                    
//!  @param [in] y        ｙ座標                                                    
//!  @param [in] pattern  パターン番号                                              
//!  @retval     true     描画                                                      
//!  @retval     false    描画範囲外                                                
// -------------------------------------------------------------------------------- 
bool RAMFUNC obj_drawPattern(int8_t x, int8_t y, uint16_t pattern)
{
	uint16_t	const *p_patData, *p_patDataEnd ;
	uint16_t	data, *p_txt, *p_txtEnd ;
	int8_t		ex, ey ;
	uint8_t		w, h, nw ;
	int			lineData, lineTxt ;


	p_patData = g_p_patData[pattern] ;

	x         += ((int8_t  *)p_patData)[0] ;
	y         += ((int8_t  *)p_patData)[1] ;
	w          = ((uint8_t *)p_patData)[2] ;
	h          = ((uint8_t *)p_patData)[3] ;
	p_patData += 2 ;

	// 上下クリップ
	ey = y + h ;
	if( y < 0 ) {

		// 画面外なら無視
		if( ey <= 0 ) {
			return false ;
		}

		p_patData += (w * -y) ;
		y          = 0 ;
		h          = ey ;

	} else
	if( ey > DRAW_HEIGHT ) {

		// 画面外なら無視
		if( y >= DRAW_HEIGHT ) {
			return false ;
		}

		h = DRAW_HEIGHT - y ;

	}

	p_txt  = &g.vram.txt[y][x] ;

	// 左右クリップ
	ex = x + w ;
	if( x < 0 ) {				// 左クリップ

		// 画面外なら無視
		if( ex <= 0 ) {
			return false ;
		}

		p_patData -= x ;
		p_txt     -= x ;
		lineData   = -x ;
		lineTxt    = DRAW_WIDTH - (w + x) ;
		x          = 0 ;
		w          = ex ;

	} else
	if( ex > DRAW_WIDTH ) {	// 右クリップ

		// 画面外なら無視
		if( x >= DRAW_WIDTH ) {
			return false ;
		}

		nw       = DRAW_WIDTH - x ;
		lineData = w - nw ;
		lineTxt  = DRAW_WIDTH - nw ;
		w        = nw ;

	} else {					// クリップなし

		lineData = 0 ;
		lineTxt  = DRAW_WIDTH - w ;

	}

	// 描画
	p_txtEnd = &p_txt[h*SCREEN_WIDTH] ;

	do {
		p_patDataEnd = &p_patData[w] ;
		do {
			data = *p_patData ;
			if( data ) {

				// 重なり点滅
//				if( (*p_txt&WALL) && (g.count1&1) ) {
//					*p_txt = data ;
//				} else {
					*p_txt = (*p_txt & WALL) | data ;
//				}
			}
			p_txt++ ;
		} while( ++p_patData != p_patDataEnd ) ;

		p_patData += lineData ;
		p_txt     += lineTxt ;
	} while( p_txt != p_txtEnd ) ;

	return true ;
}
