// -------------------------------------------------------------------------------- 
//!  @file  force.c                                                                 
//!  フォースタスクのソースファイル                                                 
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "force.h"
#include "rlaser.h"


// -------------------------------------------------------------------------------- 
//!  ビームタスクを破棄                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC beam_destroy( TCB *p_tcb )
{
	FORCEBEAM	*p_forceBeam ;


	p_forceBeam = (FORCEBEAM *)p_tcb->p_address ;
	p_forceBeam->remain++ ;

	task_detach( p_tcb ) ;
	task_releaseTcb( p_tcb ) ;
}


// -------------------------------------------------------------------------------- 
//!  ビームタスク処理                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC beam( TCB *p_tcb )
{
	int8_t	x, y ;

	// 移動
	p_tcb->bx.n += (getCos( p_tcb->angle ).n * toFp( 0.6 ) / FP_ONE) ;
	p_tcb->by.n += (getSin( p_tcb->angle ).n * toFp( 0.6 ) / FP_ONE) ;

	// 画面外
	x = p_tcb->bx.h - g.bg.x.h ;
	y = p_tcb->by.h - g.bg.y.h ;
	if( ((uint8_t )x>=DRAW_WIDTH) || ((uint8_t )y+1>=(DRAW_HEIGHT+1)) ) {
		beam_destroy( p_tcb ) ;
		return ;
	}

	// 壁
	if( g.vram.txt[y][x] & WALL ) {
		beam_destroy( p_tcb ) ;
		return ;
	}

	g.vram.txt[y][x] = 0x5000 | p_tcb->pattern ;
}


// -------------------------------------------------------------------------------- 
//!  ビームタスクを生成                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  親 TCB                                                  
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC beam_create( TCB *p_parent )
{
	TCB			*p_tcb ;
	FORCEBEAM	*p_forceBeam ;
	uint32_t	angle, pattern ;
	int			index, bx, by ;


	if( !(g.pad.once&PAD_BTN( 0 )) ) {
		return ;
	}

	if( p_parent->status & TCB_STATUS_LEFT ) {
		angle   = ( 136<<24) | ( 192<<16) | ( 120<<8) | (  64<<0) ;		// 左上・上・左下・下
		pattern = (0x80<<24) | (0x88<<16) | (0x94<<8) | (0x88<<0) ;		// 左上・上・左下・下
	} else {
		angle   = ( 192<<24) | ( 248<<16) | (  64<<8) | (   8<<0) ;		// 上・右上・下・右下
		pattern = (0x88<<24) | (0x80<<16) | (0x88<<8) | (0x94<<0) ;		// 上・右上・下・右下
	}

	bx = p_parent->x.h + g.bg.x.h ;
	by = p_parent->y.h + g.bg.y.h - 1 ;

	p_forceBeam = &g.forceBeam[0] ;
	for(index=0; index<4; index++,p_forceBeam++,angle<<=8,pattern<<=8) {
		if( !p_forceBeam->remain ) {
			continue ;
		}
		p_forceBeam->remain-- ;

		p_tcb            = task_getTcb() ;
		p_tcb->p_address = p_forceBeam ;
		p_tcb->taskProc  = beam ;
		p_tcb->x.n       = toFp( bx ) ;
		p_tcb->y.n       = toFp( by + ((index / 2) & 1) ) + toFp( 0.5 ) ;
		p_tcb->pattern   = pattern >> 24 ;
		p_tcb->angle     = angle   >> 24 ;

		task_attachHead(&g.playerList, p_tcb) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  フォースとプレイヤーの当たりチェック                                           
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @retval         true   当たり                                                  
//!  @retval         false  当たりなし                                              
// -------------------------------------------------------------------------------- 
static bool RAMFUNC force_hitCheck( TCB *p_tcb )
{
	int8_t	x, y ;


	y = p_tcb->y.h - g.playerTcb.y.h ;
	if( (uint )(y+1) > 2 ) {
		return false ;
	}

	x = p_tcb->x.h - g.playerTcb.x.h ;
	if( (uint )(x+2) > 4 ) {
		return false ;
	}

	p_tcb->status &= ~TCB_STATUS_WALL ;

	if( x < 0 ) {	// 左へ
		p_tcb->status |= (TCB_STATUS_ATTACH | TCB_STATUS_LEFT) ;

		p_tcb->xofs = -4 ;
		obj_setAction(p_tcb, ACT_FORCEL_L) ;

	} else {		// 右へ
		p_tcb->status &= ~TCB_STATUS_LEFT ;
		p_tcb->status |=  TCB_STATUS_ATTACH ;

		p_tcb->xofs =  4 ;
		obj_setAction(p_tcb, ACT_FORCEL_R) ;
	}

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  フォースタスク処理                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC force_movePos( TCB *p_tcb ) ;

static void RAMFUNC force_fire( TCB *p_tcb )
{
	uint16_t	*p_txt ;
	int			add ;


	do {
		// スクロール
		if( g.bg.xofs ) {
			p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h] ;
			if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
				p_tcb->x.h     -= g.bg.xofs ;
				p_tcb->taskProc = force_movePos ;
				break ;
			}
		}

		// 画面端
		if( p_tcb->status & TCB_STATUS_LEFT ) {

			if( p_tcb->x.h <= 0 ) {
				p_tcb->x.n      = toFp( 0 ) ;
				p_tcb->taskProc = force_movePos ;
				break ;
			}
			p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h-2] ;
			add   = -1 ;

		} else {

			if( p_tcb->x.h >= (DRAW_WIDTH-1) ) {
				p_tcb->x.n      = toFp( DRAW_WIDTH-1 ) ;
				p_tcb->taskProc = force_movePos ;
				break ;
			}
			p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h+1] ;
			add   = 1 ;

		}

		// 壁
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->taskProc = force_movePos ;
			break ;
		}

		p_tcb->x.h += add ;
	} while( 0 ) ;

	obj_updateAction( p_tcb ) ;
	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, p_tcb->pattern) ;

	// ビーム生成
	beam_create( p_tcb ) ;
}

static void RAMFUNC force_attach( TCB *p_tcb )
{
	// 移動
	p_tcb->x.h = g.playerTcb.x.h + p_tcb->xofs ;
	p_tcb->y.h = g.playerTcb.y.h ;

	obj_updateAction( p_tcb ) ;
	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, p_tcb->pattern) ;

	// 反射レーザー生成
	rlaser_create( p_tcb ) ;

	// 発射
	if( g.pad.once & PAD_BTN( 1 ) ) {
		p_tcb->status  &= ~TCB_STATUS_ATTACH ;
		p_tcb->taskProc = force_fire ;
		return ;
	}
}


static void RAMFUNC force_movePlayer( TCB *p_tcb )
{
	uint8_t		angle ;
	int8_t		x, y ;
	uint16_t	*p_txt ;

	// スクロール
	if( g.bg.xofs ) {
		p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h] ;
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->x.h -= g.bg.xofs ;
		}
	}

	// 角度取得
	angle = getAngle(p_tcb->x.h, p_tcb->y.h, g.playerTcb.x.h, g.playerTcb.y.h) ;

	// ｘ方向
	x           = p_tcb->x.h ;
	p_tcb->x.n += (getCos( angle ).n * toFp( 0.2 ) / FP_ONE) ;

	if( p_tcb->x.h != x ) {
		if( p_tcb->x.h > x ) {		// 右へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h] ;
		} else {					// 左へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h-1] ;
		}

		// 壁
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->x.n = toFp( x ) ;
		}
	}

	// ｙ方向
	y           = p_tcb->y.h ;
	p_tcb->y.n += (getSin( angle ).n * toFp( 0.2 ) / FP_ONE) ;

	if( p_tcb->y.h != y ) {
		if( p_tcb->y.h > y ) {		// 下へ
			p_txt = &g.vram.txt[p_tcb->y.h][0] ;
		} else {					// 上へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][0] ;
		}

		// 壁
		if( (p_txt[p_tcb->x.h-1]|p_txt[p_tcb->x.h]) & WALL ) {
			p_tcb->y.n = toFp( y ) ;
		}
	}

	if( p_tcb->x.n <= toFp( 1 ) ) {
		p_tcb->x.n  = toFp( 1 ) ;
	}

	obj_updateAction( p_tcb ) ;
	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, p_tcb->pattern) ;

	// ビーム生成
	beam_create( p_tcb ) ;

	// フォースとプレイヤー当たりチェック
	if( force_hitCheck( p_tcb ) != false ) {
		p_tcb->taskProc = force_attach ;
		return ;
	}
}


static void RAMFUNC force_movePos( TCB *p_tcb )
{
	FI8			nx, ny ;
	int8_t		x ;
	uint16_t	*p_txt ;

	// スクロール
	if( g.bg.xofs ) {
		p_tcb->status &= ~TCB_STATUS_WALL ;

		p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h] ;
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->status |= TCB_STATUS_WALL ;
			p_tcb->x.h    -= g.bg.xofs ;
		}
	}

	// 目標ｘ座標
	if( g.playerTcb.x.h <= (DRAW_WIDTH/2) ) {
		x = (DRAW_WIDTH / 8) * 6 ;
	} else {
		x = (DRAW_WIDTH / 8) * 2 ;
	}

	// ｘ方向
	nx.n = p_tcb->x.n ;
	if( p_tcb->x.h < x ) {						// 右へ
		p_tcb->x.n += toFp( 0.1 ) ;
		if( p_tcb->x.h != nx.h ) {
			p_tcb->status &= ~TCB_STATUS_WALL ;
			p_txt          = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h] ;
			if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
				p_tcb->status |= TCB_STATUS_WALL ;
				p_tcb->x.n     = toFp( nx.h ) ;
			}
		}
	} else
	if( p_tcb->x.h > x ) {						// 左へ
		p_tcb->x.n -= toFp( 0.1 ) ;
		if( p_tcb->x.h != nx.h ) {
			p_tcb->status &= ~TCB_STATUS_WALL ;
			p_txt          = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h-1] ;
			if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
				p_tcb->status |= TCB_STATUS_WALL ;
				p_tcb->x.n     = toFp( nx.h ) ;
			}
		}
	}

	// ｙ方向
	if( (p_tcb->x.h==x) || (p_tcb->status&TCB_STATUS_WALL) ) {
		ny.n = p_tcb->y.n ;

		if( p_tcb->y.h < g.playerTcb.y.h ) {	// 下へ
			p_tcb->y.n += toFp( 0.1 ) ;
			if( p_tcb->y.h != ny.h ) {
				p_txt = &g.vram.txt[p_tcb->y.h][0] ;
				if( (p_txt[p_tcb->x.h-1]|p_txt[p_tcb->x.h]) & WALL ) {
					p_tcb->y.n = toFp( ny.h ) ;
				}
			}
		} else
		if( p_tcb->y.h > g.playerTcb.y.h ) {	// 上へ
			p_tcb->y.n -= toFp( 0.1 ) ;
			if( p_tcb->y.h != ny.h ) {
				p_txt = &g.vram.txt[p_tcb->y.h-1][0] ;
				if( (p_txt[p_tcb->x.h-1]|p_txt[p_tcb->x.h]) & WALL ) {
					p_tcb->y.n = toFp( ny.h ) ;
				}
			}
		}
	}

	if( p_tcb->x.n <= toFp( 1 ) ) {
		p_tcb->x.n  = toFp( 1 ) ;
	}

	obj_updateAction( p_tcb ) ;
	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, p_tcb->pattern) ;

	// ビーム生成
	beam_create( p_tcb ) ;

	// フォースとプレイヤー当たりチェック
	if( force_hitCheck( p_tcb ) != false ) {
		p_tcb->taskProc = force_attach ;
		return ;
	}

	// プレイヤーヘ移動
	if( g.pad.once & PAD_BTN( 1 ) ) {
		p_tcb->taskProc = force_movePlayer ;
	}
}


// -------------------------------------------------------------------------------- 
//!  フォースタスクを生成                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC force_create( void )
{
	TCB		*p_tcb ;


	g.forceBeam[0].remain = 8 ;
	g.forceBeam[1].remain = 8 ;
	g.forceBeam[2].remain = 8 ;
	g.forceBeam[3].remain = 8 ;

	p_tcb           = &g.forceTcb ;
	p_tcb->taskProc = force_movePos ;
	p_tcb->x.n      = toFp( DRAW_WIDTH + 1 - 10 ) ;
	p_tcb->y.n      = toFp( DRAW_HEIGHT/2 ) ;

	obj_setAction(&g.forceTcb, ACT_FORCEL_R) ;

	task_attachNext(&g.playerTcb, p_tcb) ;
}
