// -------------------------------------------------------------------------------- 
//!  @file  player.c                                                                
//!  プレイヤータスクのソースファイル                                               
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "player.h"


	enum {
		BEAM_MAX					=	3							//! ビーム最大数
	} ;


// -------------------------------------------------------------------------------- 
//!  ビームタスクを破棄                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC beam_destroy( TCB *p_tcb )
{
	p_tcb->p_parent->count1++ ;

	task_detach( p_tcb ) ;
	task_releaseTcb( p_tcb ) ;
}


// -------------------------------------------------------------------------------- 
//!  ビームタスク処理                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC beam( TCB *p_tcb )
{
	uint16_t	*p_txt ;
	int8_t		x, y ;

	// 移動
	p_tcb->bx.h++ ;

	// 壁
	x     = p_tcb->bx.h - g.bg.x.h ;
	y     = p_tcb->by.h - g.bg.y.h ;
	p_txt = &g.vram.txt[y-1][x] ;
	if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
		beam_destroy( p_tcb ) ;
		return ;
	}

	// 描画
	if( obj_drawPattern(x, p_tcb->y.h, p_tcb->pattern) == false ) {
		beam_destroy( p_tcb ) ;
		return ;
	}
}


// -------------------------------------------------------------------------------- 
//!  波動砲タスクを破棄                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC beamWave_destroy( TCB *p_tcb )
{
	p_tcb->p_parent->p_child = p_tcb ;

	task_detach( p_tcb ) ;
}


// -------------------------------------------------------------------------------- 
//!  波動砲タスク処理                                                               
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC beamWave( TCB *p_tcb )
{
	uint16_t	*p_txt ;
	int8_t		x, y ;

	// 移動
	p_tcb->bx.n += toFp( 0.8 ) ;

	// 壁
	x = p_tcb->bx.h - g.bg.x.h ;
	y = p_tcb->by.h - g.bg.y.h ;
	if( x < DRAW_WIDTH ) {
		p_txt = &g.vram.txt[y-1][x] ;
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			beamWave_destroy( p_tcb ) ;
			return ;
		}
	}

	obj_updateAction( p_tcb ) ;

	// 描画
	if( obj_drawPattern(x, p_tcb->y.h, p_tcb->pattern) == false ) {
		beamWave_destroy( p_tcb ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ジェットタスク処理                                                             
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void RAMFUNC jet( TCB *p_tcb )
{
	if( p_tcb->status & TCB_STATUS_ACTEND ) {
		task_detach( p_tcb ) ;
		task_releaseTcb( p_tcb ) ;
		return ;
	}

	p_tcb->x = p_tcb->p_parent->x ;
	p_tcb->y = p_tcb->p_parent->y ;

	obj_updateAction( p_tcb ) ;

	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, p_tcb->pattern) ;
}


// -------------------------------------------------------------------------------- 
//!  プレイヤータスクを生成                                                         
//!                                                                                 
//!  @param [in] p_player  プレイヤー TCB                                           
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void jet_create( TCB *p_player )
{
	TCB		*p_tcb ;


	p_tcb           = task_getTcb() ;
	p_tcb->p_parent = p_player ;
	p_tcb->taskProc = jet ;

	obj_setAction(p_tcb, ACT_PLAYERJET) ;

	task_attachNext(p_player, p_tcb) ;
}


// -------------------------------------------------------------------------------- 
//!  プレイヤー移動                                                                 
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static inline void RAMFUNC player_move( TCB *p_tcb )
{
	static const uint8_t	dirAngle[]={
								1,				//
								128,			// 左
								0,				// 右
								1,				//
								192,			// 上
								160,			// 左上
								224,			// 右上
								1,				//
								64,				// 下
								96,				// 左下
								32,				// 右下
								1,				//
								1,				// 上
								1,				// 左上
								1,				// 右上
								1				//
							} ;
	int8_t					x, y ;
	uint8_t					angle ;
	uint16_t				*p_txt ;

	// 移動
	angle = dirAngle[(g.pad.norm>>PAD_DIR_)&15] ;

	// 入力なし
	if( angle == 1 ) {
		return ;
	}

	// ｘ方向
	x           = p_tcb->x.h ;
	p_tcb->x.n += (getCos( angle ).n * p_tcb->speed / 8) ;

	if( x != p_tcb->x.h ) {
		if( x > p_tcb->x.h ) {		// 左へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][x-3] ;
		} else {					// 右へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][x+2] ;
		}

		// 壁
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->x.n = toFp( x ) ;
		}

		if( p_tcb->x.n < toFp( 5 ) ) {
			p_tcb->x.n = toFp( 5 ) ;
		} else
		if( p_tcb->x.n > toFp( DRAW_WIDTH-5 ) ) {
			p_tcb->x.n = toFp( DRAW_WIDTH-5 ) ;
		}
	}

	// ｙ方向
	y           = p_tcb->y.h ;
	p_tcb->y.n += (getSin( angle ).n * p_tcb->speed / 8) ;

	if( y != p_tcb->y.h ) {
		if( y > p_tcb->y.h ) {		// 上へ
			p_txt = &g.vram.txt[p_tcb->y.h-1][x-2] ;
		} else {					// 下へ
			p_txt = &g.vram.txt[p_tcb->y.h-0][x-2] ;
		}

		// 壁
		if( (p_txt[0]|p_txt[1]|p_txt[2]|p_txt[3]) & WALL ) {
			p_tcb->y.n = toFp( y ) ;
		}

		if( p_tcb->y.n < toFp( 1 ) ) {
			p_tcb->y.n = toFp( 1 ) ;
		} else
		if( p_tcb->y.n > toFp( DRAW_HEIGHT-1 ) ) {
			p_tcb->y.n = toFp( DRAW_HEIGHT-1 ) ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  プレイヤータスク処理                                                           
//!                                                                                 
//!  @param [in,out] p_tcb  タスクコントロールブロック                              
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void RAMFUNC player( TCB *p_tcb )
{
	TCB			*p_beam ;
	uint8_t		action ;
	uint16_t	*p_txt ;


#ifdef  DEBUG
	if( g.pad.once & PAD_RB ) {
		p_tcb->speed = 1 + (p_tcb->speed & 3) ;
		jet_create( p_tcb ) ;
	}
#endif//DEBUG

	// スクロール
	if( (p_tcb->x.n>toFp( 5 )) && g.bg.xofs ) {
		p_txt = &g.vram.txt[p_tcb->y.h-1][p_tcb->x.h+g.bg.xofs] ;
		if( (p_txt[0]|p_txt[SCREEN_WIDTH]) & WALL ) {
			p_tcb->x.h -= g.bg.xofs ;
		}
	}

	// 移動
	player_move( p_tcb ) ;

	obj_drawPattern(p_tcb->x.h, p_tcb->y.h, PAT_PLAYER) ;

	// ビーム
	if( g.pad.once & PAD_BTN( 0 ) ) {
		if( p_tcb->count1 ) {
			p_beam           = task_getTcb() ;
			p_beam->p_parent = p_tcb ;
			p_beam->taskProc = beam ;
			p_beam->pattern  = PAT_BEAM ;
			p_beam->bx.n     = g.bg.x.n + p_tcb->x.n ;
			p_beam->by.n     = g.bg.y.n + p_tcb->y.n ;

			task_attachPrev(p_tcb, p_beam) ;

			p_tcb->count1-- ;
		}
	}

	// 波動砲
	if( g.pad.norm & PAD_BTN( 0 ) ) {
		if( p_tcb->count0 != 80 ) {
			p_tcb->count0++ ;
		}
	} else {
		do {
			// 溜め不足なら無視
			if( p_tcb->count0 <= 10 ) {
				break ;
			}

			// 発射中なら無視
			if( p_tcb->p_child == NULL ) {
				break ;
			}

			action = ACT_BEAMWAVE0 ;
			if( p_tcb->count0 == 80 ) {
				action = ACT_BEAMWAVE3 ;
			} else
			if( p_tcb->count0 >= 60 ) {
				action = ACT_BEAMWAVE2 ;
			} else
			if( p_tcb->count0 >= 35 ) {
				action = ACT_BEAMWAVE1 ;
			}

			p_beam           = p_tcb->p_child ;
			p_beam->p_parent = p_tcb ;
			p_beam->taskProc = beamWave ;
			p_beam->bx.n     = g.bg.x.n + p_tcb->x.n ;
			p_beam->by.n     = g.bg.y.n + p_tcb->y.n ;

			obj_setAction(p_beam, action) ;

			task_attachPrev(p_tcb, p_beam) ;

			// フォース右あり
			if( (g.forceTcb.status&(TCB_STATUS_ATTACH|TCB_STATUS_LEFT)) == TCB_STATUS_ATTACH ) {
				p_beam->bx.h += 2 ;
			}

			p_tcb->p_child = NULL ;
		} while( 0 ) ;

		p_tcb->count0 = 0 ;
	}
}


// -------------------------------------------------------------------------------- 
//!  プレイヤータスクを生成                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void player_create( void )
{
	TCB		*p_tcb ;


	p_tcb           = &g.playerTcb ;
	p_tcb->p_child  = task_getTcb() ;
	p_tcb->taskProc = player ;
	p_tcb->x.n      = toFp( 10 ) ;
	p_tcb->y.n      = toFp( DRAW_HEIGHT/2 ) ;
	p_tcb->speed    = 1 ;
	p_tcb->count1   = BEAM_MAX ;

	task_attachTail(&g.playerList, p_tcb) ;
}
