// -------------------------------------------------------------------------------- 
//!  @file  matrix.c                                                                
//!  行列演算のソースファイル                                                       
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "matrix.h"


// -------------------------------------------------------------------------------- 
//!  ｘ軸回の行列を作成                                                             
//!                                                                                 
//!  @param [out] p_matrix  行列                                                    
//!  @param [in]  angle     角度                                                    
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC matrix_rotationX(MATRIX *p_matrix, uint8_t angle)
{
	p_matrix->m[2][0] = 0 ;
	p_matrix->m[2][1] = getSin( angle ) ;
	p_matrix->m[2][2] = getCos( angle ) ;

	p_matrix->m[1][0] = 0 ;
	p_matrix->m[1][1] =  p_matrix->m[2][2] ;
	p_matrix->m[1][2] = -p_matrix->m[2][1] ;

	p_matrix->m[0][0] = FP_ONE ;
	p_matrix->m[0][1] = 0 ;
	p_matrix->m[0][2] = 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ｙ軸回の行列を作成                                                             
//!                                                                                 
//!  @param [out] p_matrix  行列                                                    
//!  @param [in]  angle     角度                                                    
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC matrix_rotationY(MATRIX *p_matrix, uint8_t angle)
{
	p_matrix->m[0][0] = getCos( angle ) ;
	p_matrix->m[0][1] = 0 ;
	p_matrix->m[0][2] = getSin( angle ) ;

	p_matrix->m[2][0] = -p_matrix->m[0][2] ;
	p_matrix->m[2][1] = 0 ;
	p_matrix->m[2][2] =  p_matrix->m[0][0] ;

	p_matrix->m[1][0] = 0 ;
	p_matrix->m[1][1] = FP_ONE ;
	p_matrix->m[1][2] = 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ｚ軸回の行列を作成                                                             
//!                                                                                 
//!  @param [out] p_matrix  行列                                                    
//!  @param [in]  angle     角度                                                    
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC matrix_rotationZ(MATRIX *p_matrix, uint8_t angle)
{
	p_matrix->m[1][0] = getSin( angle ) ;
	p_matrix->m[1][1] = getCos( angle ) ;
	p_matrix->m[1][2] = 0 ;

	p_matrix->m[0][0] =  p_matrix->m[1][1] ;
	p_matrix->m[0][1] = -p_matrix->m[1][0] ;
	p_matrix->m[0][2] = 0 ;

	p_matrix->m[2][0] = 0 ;
	p_matrix->m[2][1] = 0 ;
	p_matrix->m[2][2] = FP_ONE ;
}


// -------------------------------------------------------------------------------- 
//!  行列の乗算                                                                     
//!                                                                                 
//!  @param [in]  p_m0      被乗数                                                  
//!  @param [in]  p_m1      乗数                                                    
//!  @param [out] p_result  演算結果格納アドレス                                    
//!                         （p_m0 と同じアドレスを指定可能）                       
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC matrix_mul(MATRIX *p_m0, MATRIX *p_m1, MATRIX *p_result)
{
	int24f8_t	m[3] ;


	m[0]              = ((p_m0->m[0][0] * p_m1->m[0][0]) + (p_m0->m[0][1] * p_m1->m[0][1]) + (p_m0->m[0][2] * p_m1->m[0][2])) / FP_ONE ;
	m[1]              = ((p_m0->m[0][0] * p_m1->m[1][0]) + (p_m0->m[0][1] * p_m1->m[1][1]) + (p_m0->m[0][2] * p_m1->m[1][2])) / FP_ONE ;
	m[2]              = ((p_m0->m[0][0] * p_m1->m[2][0]) + (p_m0->m[0][1] * p_m1->m[2][1]) + (p_m0->m[0][2] * p_m1->m[2][2])) / FP_ONE ;
	p_result->m[0][0] = m[0] ;
	p_result->m[0][1] = m[1] ;
	p_result->m[0][2] = m[2] ;

	m[0]              = ((p_m0->m[1][0] * p_m1->m[0][0]) + (p_m0->m[1][1] * p_m1->m[0][1]) + (p_m0->m[1][2] * p_m1->m[0][2])) / FP_ONE ;
	m[1]              = ((p_m0->m[1][0] * p_m1->m[1][0]) + (p_m0->m[1][1] * p_m1->m[1][1]) + (p_m0->m[1][2] * p_m1->m[1][2])) / FP_ONE ;
	m[2]              = ((p_m0->m[1][0] * p_m1->m[2][0]) + (p_m0->m[1][1] * p_m1->m[2][1]) + (p_m0->m[1][2] * p_m1->m[2][2])) / FP_ONE ;
	p_result->m[1][0] = m[0] ;
	p_result->m[1][1] = m[1] ;
	p_result->m[1][2] = m[2] ;

	m[0]              = ((p_m0->m[2][0] * p_m1->m[0][0]) + (p_m0->m[2][1] * p_m1->m[0][1]) + (p_m0->m[2][2] * p_m1->m[0][2])) / FP_ONE ;
	m[1]              = ((p_m0->m[2][0] * p_m1->m[1][0]) + (p_m0->m[2][1] * p_m1->m[1][1]) + (p_m0->m[2][2] * p_m1->m[1][2])) / FP_ONE ;
	m[2]              = ((p_m0->m[2][0] * p_m1->m[2][0]) + (p_m0->m[2][1] * p_m1->m[2][1]) + (p_m0->m[2][2] * p_m1->m[2][2])) / FP_ONE ;
	p_result->m[2][0] = m[0] ;
	p_result->m[2][1] = m[1] ;
	p_result->m[2][2] = m[2] ;
}


// -------------------------------------------------------------------------------- 
//!  行列からベクトルを変換                                                         
//!                                                                                 
//!  @param [in]  p_matrix  行列                                                    
//!  @param [in]  p_vector  ベクトル                                                
//!  @param [out] p_result  演算結果格納アドレス                                    
//!                         （p_vector と同じアドレスを指定可能）                   
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC matrix_trans(MATRIX *p_matrix, VECTOR3 *p_vector, VECTOR3 *p_result)
{
	VECTOR3		vector ;


	vector.x = ((p_matrix->m[0][0] * p_vector->x) + (p_matrix->m[0][1] * p_vector->y) + (p_matrix->m[0][2] * p_vector->z)) / FP_ONE ;
	vector.y = ((p_matrix->m[1][0] * p_vector->x) + (p_matrix->m[1][1] * p_vector->y) + (p_matrix->m[1][2] * p_vector->z)) / FP_ONE ;
	vector.z = ((p_matrix->m[2][0] * p_vector->x) + (p_matrix->m[2][1] * p_vector->y) + (p_matrix->m[2][2] * p_vector->z)) / FP_ONE ;

	*p_result = vector ;
}
