// -------------------------------------------------------------------------------- 
//!  @file  polygon.c                                                               
//!  ポリゴン描画のソースファイル                                                   
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include "common.h"
#include "hardware/divider.h"
#include "polygon.h"

	//! 最小値・最大値
	typedef struct {
		int16_t						min ;							//!< 最小値
		int16_t						max ;							//!< 最大値
	} MINMAX ;


static MINMAX						m_minMax[SCREEN_HEIGHT] ;		//!< 最小値・最大値


// -------------------------------------------------------------------------------- 
//!  反時計回りか調べる                                                             
//!                                                                                 
//!  @param [in] p_point  三角形の点位置                                            
//!  @retval     true     反時計回り                                                
//!  @retval     true     時計回りか位置異常                                        
// -------------------------------------------------------------------------------- 
static bool RAMFUNC isCounterClockwise( POINT16 *p_point )
{
	return (bool )((((p_point[1].x - p_point[0].x) * (p_point[2].y - p_point[1].y)) -
	                ((p_point[2].x - p_point[1].x) * (p_point[1].y - p_point[0].y))) < 0) ;
}


// -------------------------------------------------------------------------------- 
//!  最小値・最大値を設定                                                           
//!                                                                                 
//!  @param [in] sx  開始ｘ座標                                                     
//!  @param [in] sy  開始ｘ座標                                                     
//!  @param [in] ex  終了ｘ座標                                                     
//!  @param [in] ey  終了ｘ座標                                                     
//!  @return         なし                                                           
// -------------------------------------------------------------------------------- 
static void RAMFUNC setMinMax(int sx, int sy, int ex, int ey)
{
	MINMAX		*p_minMax ;
	uint8_t		clip[2] ;
	int			width, height, addx, addy, value ;

	// 開始ｙ座標
	clip[0] = 0 ;
	if( sy < 0 ) {
		clip[0] = 1 ;
	} else
	if( sy >= SCREEN_HEIGHT ) {
		clip[0] = 2 ;
	}

	// 終了ｙ座標
	clip[1] = 0 ;
	if( ey < 0 ) {
		clip[1] = 1 ;
	} else
	if( ey >= SCREEN_HEIGHT ) {
		clip[1] = 2 ;
	}

	width  = (ex - sx) ;
	height = (ey - sy) ;

	// クリップあり
	if( clip[0] | clip[1] ) {

		// 表示範囲外
		if( clip[0] & clip[1] ) {
			return ;
		}

		// 開始クリップ
		if( clip[0] ) {
			if( clip[0] == 1 ) {	// 上クリップ
				hw_divider_divmod_s32_start(width*(              0-sy)*16, height) ;
				sy = 0 ;
			} else {				// 下クリップ
				hw_divider_divmod_s32_start(width*(SCREEN_HEIGHT-1-sy)*16, height) ;
				sy = SCREEN_HEIGHT - 1 ;
			}

			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
//#			__asm volatile ("nop") ;
//#			__asm volatile ("nop") ;

			sx = (int32_t )sio_hw->div_quotient / 16 + sx ;
		}

		// 終了クリップ
		if( clip[1] ) {
			if( clip[1] == 1 ) {	// 上クリップ
				hw_divider_divmod_s32_start(width*(              0-sy)*16, height) ;
				ey = 0 ;
			} else {				// 下クリップ
				hw_divider_divmod_s32_start(width*(SCREEN_HEIGHT-1-sy)*16, height) ;
				ey = SCREEN_HEIGHT - 1 ;
			}

			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
			__asm volatile ("nop") ;
//#			__asm volatile ("nop") ;
//#			__asm volatile ("nop") ;

			ex = (int32_t )sio_hw->div_quotient / 16 + sx ;
		}
	}

	addx = 1 ;
	if( width < 0 ) {
		width = -width ;
		addx  = -1 ;
	}

	addy = 1 ;
	if( height < 0 ) {
		height = -height ;
		addy   = -1 ;
	}

	if( width >= height ) {

		// 方向を左から右へ
		if( addx < 0 ) {
			value = sx ;
			sx    = ex ;
			ex    = value ;

			sy    = ey ;
			addy  = -addy ;
		}

		p_minMax = &m_minMax[sy] ;

		value = width / 2 ;

#if 1
		if( p_minMax->min > sx ) {
			p_minMax->min = sx ;
		}

		while( 1 ) {
			value -= height ;
			if( value < 0 ) {

				if( p_minMax->max < sx ) {
					p_minMax->max = sx ;
				}

				if( ++sx > ex ) {
					break ;
				}

				value    += width ;
				p_minMax += addy ;

				if( p_minMax->min > sx ) {
					p_minMax->min = sx ;
				}

			} else {

				if( ++sx > ex ) {
					if( p_minMax->max < sx ) {
						p_minMax->max = sx ;
					}
					break ;
				}

			}
		}
#else
		while( sx <= ex ) {
			if( p_minMax->min > sx ) {
				p_minMax->min = sx ;
			}

			if( p_minMax->max < sx ) {
				p_minMax->max = sx ;
			}

			value -= height ;
			if( value < 0 ) {
				value    += width ;
				p_minMax += addy ;
			}

			sx++ ;
		}
#endif

	} else {

		// 方向を上から下へ
		if( addy < 0 ) {
			sx   = ex ;
			addx = -addx ;

			value = sy ;
			sy    = ey ;
			ey    = value ;
		}

		p_minMax = &m_minMax[sy] ;

		value = height / 2 ;

		while( sy <= ey ) {
			if( p_minMax->min > sx ) {
				p_minMax->min = sx ;
			}

			if( p_minMax->max < sx ) {
				p_minMax->max = sx ;
			}

			value -= width ;
			if( value < 0 ) {
				value  += height ;
				sx     += addx ;
			}

			p_minMax++ ;
			sy++ ;
		}

	}
}


#if     (BITMAP_MODE==4)
// -------------------------------------------------------------------------------- 
//!  最小値・最大値の間を塗り潰す（8bit）                                           
//!                                                                                 
//!  @param [in] min       最小値                                                   
//!  @param [in] max       最大値                                                   
//!  @param [in] pattern   色パターン                                               
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void RAMFUNC fillMinMax(int min, int max, uint32_t pattern)
{
	MINMAX				*p_minMax, *p_minMaxEnd ;
	volatile uint32_t	value ;
	uint32_t			mask, *p_bitmap, *p_dest ;
	int					count, sx, ex, width ;


	if( min & 1 ) {
		pattern = (pattern << 4) | (pattern >> 28) ;
	}

	p_minMax    = &m_minMax[min] ;
	p_minMaxEnd = &m_minMax[max] ;
	p_bitmap    = (uint32_t *)&g.p_bitmap[min*(SCREEN_WIDTH/2)] ;

	do {
		sx = p_minMax->min ;
		if( sx < 0 ) {
			sx = 0 ;
		}

		ex = p_minMax->max ;
		if( ex > (SCREEN_WIDTH-1) ) {
			ex = (SCREEN_WIDTH-1) ;
		}

		width  = ex - sx + 1 ;
		p_dest = &p_bitmap[sx/8] ;

		sx &= 7 ;

		if( (sx+width) <= 8 ) {		// ４バイト未満

			mask    = (0xffffffff >> ((8 - width) * 4)) << (sx * 4) ;
			*p_dest = (*p_dest & ~mask) | (pattern & mask) ;

		} else {					// ４バイト以上

			// 開始端数
			if( sx ) {
				mask    = (0xffffffff << (sx * 4)) ;
				*p_dest = (*p_dest & ~mask) | (pattern & mask) ;

				p_dest++ ;
				width -= (8 - sx) ;
			}

			count = width / 8 ;

			// 終了端数
			width &= 7 ;
			if( width ) {
				mask          = (0xffffffff >> ((8 - width) * 4)) ;
				p_dest[count] = (p_dest[count] & ~mask) | (pattern & mask) ;
			}

			// 中塗り潰し
			if( count ) {
				dma_channel_wait_for_finish_blocking( DMACH_POLYGON ) ;
				value = pattern ;
				dma_channel_configure(DMACH_POLYGON, &g.dmaConfigPolygon, &p_dest[0], &value, count, true) ;
			}

		}

		pattern = (pattern << 4) | (pattern >> 28) ;

		p_bitmap += (SCREEN_WIDTH / 8) ;
	} while( ++p_minMax <= p_minMaxEnd ) ;

	dma_channel_wait_for_finish_blocking( DMACH_POLYGON ) ;
}
#else
// -------------------------------------------------------------------------------- 
//!  最小値・最大値の間を塗り潰す（8bit）                                           
//!                                                                                 
//!  @param [in] min       最小値                                                   
//!  @param [in] max       最大値                                                   
//!  @param [in] pattern   色パターン                                               
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void RAMFUNC fillMinMax(int min, int max, uint32_t pattern)
{
	MINMAX		*p_minMax, *p_minMaxEnd ;
	uint32_t	mask, *p_bitmap, *p_dest ;
	int			count, sx, ex, width ;


	p_minMax    = &m_minMax[min] ;
	p_minMaxEnd = &m_minMax[max] ;
	p_bitmap    = (uint32_t *)&g.p_bitmap[min*SCREEN_WIDTH] ;

	do {
		sx = p_minMax->min ;
		if( sx < 0 ) {
			sx = 0 ;
		}

		ex = p_minMax->max ;
		if( ex > (SCREEN_WIDTH-1) ) {
			ex = (SCREEN_WIDTH-1) ;
		}

		width  = ex - sx + 1 ;
		p_dest = &p_bitmap[sx/4] ;

		sx &= 3 ;

		if( (sx+width) <= 4 ) {		// ４バイト未満

			mask    = (0xffffffff >> ((4 - width) * 8)) << (sx * 8) ;
			*p_dest = (*p_dest & ~mask) | (pattern & mask) ;

		} else {					// ４バイト以上

			// 開始端数
			if( sx ) {
				mask    = (0xffffffff << (sx * 8)) ;
				*p_dest = (*p_dest & ~mask) | (pattern & mask) ;

				p_dest++ ;
				width -= (4 - sx) ;
			}

			count = width / 4 ;

			// 終了端数
			width &= 3 ;
			if( width ) {
				mask          = (0xffffffff >> ((4 - width) * 8)) ;
				p_dest[count] = (p_dest[count] & ~mask) | (pattern & mask) ;
			}

			// 中塗り潰し
			if( count ) {
				dma_channel_wait_for_finish_blocking( DMACH_POLYGON ) ;
				dma_channel_configure(DMACH_POLYGON, &g.dmaConfigPolygon, &p_dest[0], &pattern, count, true) ;
			}

		}

		p_bitmap += (SCREEN_WIDTH / 4) ;
	} while( ++p_minMax <= p_minMaxEnd ) ;

	dma_channel_wait_for_finish_blocking( DMACH_POLYGON ) ;
}
#endif//(BITMAP_MODE==4)


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC polygon_init( void )
{
	dma_channel_claim( DMACH_POLYGON ) ;
	g.dmaConfigPolygon = dma_channel_get_default_config( DMACH_POLYGON ) ;
	channel_config_set_write_increment(&g.dmaConfigPolygon, true) ;
	channel_config_set_read_increment(&g.dmaConfigPolygon, false) ;
	channel_config_set_transfer_data_size(&g.dmaConfigPolygon, DMA_SIZE_32) ;
}


// -------------------------------------------------------------------------------- 
//!  三角形を描画                                                                   
//!                                                                                 
//!  @param [in] p_point  三角形の点位置                                            
//!  @param [in] palette  パレット番号                                              
//!  @retval     true     描画あり                                                  
//!  @retval     false    描画なし                                                  
// -------------------------------------------------------------------------------- 
bool RAMFUNC polygon_triangle(POINT16 *p_point, uint8_t palette)
{
	uint32_t	pattern, value ;
	int			min, max ;

	// 反時計回りでないなら無視
	if( isCounterClockwise( p_point ) == false ) {
		return false ;
	}

	// 最小値・最大値取得
	min = p_point[0].y ;
	max = min ;

	if( min > p_point[1].y ) {
		min = p_point[1].y ;
	} else
	if( max < p_point[1].y ) {
		max = p_point[1].y ;
	}

	if( min > p_point[2].y ) {
		min = p_point[2].y ;
	} else
	if( max < p_point[2].y ) {
		max = p_point[2].y ;
	}

	if( (min>=SCREEN_HEIGHT) || (max<0) ) {
		return false ;
	}

	if( min < 0 ) {
		min = 0 ;
	}

	if( max > (SCREEN_HEIGHT-1) ) {
		max = (SCREEN_HEIGHT-1) ;
	}

	// 最小値・最大値初期化
	value = (-32768 << 16) | 32767 ;
	channel_config_set_transfer_data_size(&g.dmaConfigPolygon, DMA_SIZE_32) ;
	dma_channel_configure(DMACH_POLYGON, &g.dmaConfigPolygon, &m_minMax[min], &value, max-min+1, true) ;

	// 塗り潰しパターン
	pattern  = ((uint32_t )palette << 8) | palette ;
	pattern |= (pattern << 16) ;

	dma_channel_wait_for_finish_blocking( DMACH_POLYGON ) ;

	// 最小値・最大値設定
	setMinMax(p_point[0].x, p_point[0].y, p_point[1].x, p_point[1].y) ;
	setMinMax(p_point[1].x, p_point[1].y, p_point[2].x, p_point[2].y) ;
	setMinMax(p_point[2].x, p_point[2].y, p_point[0].x, p_point[0].y) ;

	// 最小値・最大値塗り潰し
	fillMinMax(min, max, pattern) ;

	return true ;
}
